from typing import Any

from sqlalchemy import ColumnElement
from sqlmodel import or_, and_


class ConditionOperator:
    """A utility class to easily generate common expressions"""
    def __init__(self, *values: Any):
        self.values = values

    def get_expression(self, column: ColumnElement) -> ColumnElement:
        """Builds and returns the appropriate expression.

        :param column: The column on which to evaluate
        :return: the expression
        """
        raise NotImplementedError('Must implement `get_expression` in subclass')


class GreaterThan(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column > self.values[0]


class GreaterThanEqualTo(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column >= self.values[0]


class LessThan(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column < self.values[0]


class LessThanEqualTo(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column <= self.values[0]


class Between(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        lower_bound, upper_bound = self.values
        return and_(column >= lower_bound, column <= upper_bound)


class AnyOf(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return or_(*[column == value for value in self.values])


class NoneOf(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return and_(*[column != value for value in self.values])


class IsSet(ConditionOperator):
    """Expression to filter to rows that have a value set for a specific Column"""
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return or_(column == True, and_(column != None, column != False))


class NotSet(ConditionOperator):
    """Expression to filter to rows that have no value set for a specific Column"""
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return or_(column == False, column == None)


is_set = IsSet()
not_set = NotSet()



from typing import Optional

from sqlalchemy.exc import IntegrityError
from daomodel import DAOModel


class Conflict(Exception):
    """Indicates that the store could not be updated due to an existing conflict."""
    def __init__(self, model: Optional[DAOModel] = None, msg: Optional[str] = None, error: Optional[Exception] = None):
        self.detail = msg if msg else f'Conflict for {model.__class__.doc_name()} {model}'
        self.original_error = error


class TransactionMixin:
    """A mixin that provides transaction management functionality.

    self.db must be set in the extending class before calling any methods.
    """
    @property
    def _auto_commit(self) -> bool:
        """Determines if auto-commit is enabled.

        This is based on whether the session is in transaction mode -
        if in transaction mode, auto-commit is disabled.

        :return: True if auto-commit is enabled, False otherwise
        """
        if not hasattr(self.db, '_in_transaction'):
            self.db._in_transaction = False
        return not self.db._in_transaction

    def start_transaction(self) -> None:
        """Starts a transaction by setting transaction_mode to True.

        This disables auto_commit until the transaction is committed or rolled back.
        """
        self.db._in_transaction = True

    def _end_transaction(self) -> None:
        """Resets the transaction state after a transaction is committed or rolled back."""
        self.db._in_transaction = False

    def _commit_if_not_transaction(self):
        if self._auto_commit:
            self.commit()

    def commit(self, *models_to_refresh: DAOModel) -> None:
        """Commits all pending changes to the database.

        'Pending changes' includes data changes made to models that were fetched from the database.
        Use dao.start_transaction() to avoid automatically calling this method following each insert, upsert, and remove.
        This will commit all changes within the session and is not limited to this DAO.
        Following the DB commit, DAOModels will be detached, needing to be refreshed.

        If this DAO was in transaction mode, it will be reset to auto-commit mode after committing.

        :param models_to_refresh: The DAOModels to refresh after committing
        :raises Conflict: if a unique constraint is violated
        :raises IntegrityError: if a database constraint is violated (e.g., NOT NULL, foreign key)
        """
        try:
            self.db.commit()
            for model in models_to_refresh:
                self.db.refresh(model)
            self._end_transaction()
        except IntegrityError as e:
            error_msg = str(e).lower()
            if any(text in error_msg for text in ('unique constraint', 'unique violation', 'duplicate key')):
                raise Conflict(msg=f"Unique constraint violation: {str(e)}", error=e)
            raise

    def rollback(self) -> None:
        """Reverts all pending database changes of a transaction.

        This will discard all changes that have not yet been committed.

        :raises RuntimeError: if not in transaction mode
        """
        if self._auto_commit:
            raise RuntimeError('Cannot rollback while not in transaction mode')
        self.db.rollback()
        self._end_transaction()



from typing import List

from daomodel import DAOModel
from daomodel.fields import Identifier
from daomodel.testing import TestDAOFactory


class Album(DAOModel, table=True):
    id: Identifier[int]
    list: list['Photo']
    List: List['Photo']


class Photo(DAOModel, table=True):
    id: Identifier[int]
    album: Album


def test_implicit_relationship():
    with TestDAOFactory() as daos:
        album_dao = daos[Album]
        photo_dao = daos[Photo]
        album_dao.create(1)
        album_dao.create(2)
        album_dao.create(3)
        photo_dao.create_with(id=1, album=1)
        photo_dao.create_with(id=2, album=3)
        photo_dao.create_with(id=3, album=3)

        album = album_dao.get(1)
        assert album.list == album.List == [photo_dao.get(1)]
        album = album_dao.get(2)
        assert album.list == album.List == []
        album = album_dao.get(3)
        assert album.list == album.List == [photo_dao.get(2), photo_dao.get(3)]



from sqlalchemy.orm import declared_attr

import daomodel


class DAOModel(daomodel.DAOModel):
    """A DAOModel configured to act just like an SQLModel."""

    @declared_attr
    def __tablename__(self) -> str:
        return self.__name__.lower()



from typing import Any, Generator

import pytest

from daomodel.testing import TestDAOFactory


@pytest.fixture(name='daos')
def daos_fixture() -> Generator[TestDAOFactory, Any, None]:
    """
    Provides a DAOFactory for Testing as a pytest fixture named `daos`.
    """
    with TestDAOFactory() as daos:
        yield daos



import inspect
from dataclasses import dataclass
from typing import Any, Callable, TypeVar, Generic

import pytest
from _pytest.outcomes import fail
from sqlalchemy.orm import sessionmaker

from daomodel import DAOModel
from daomodel.dao import NotFound, DAO
from daomodel.db import DAOFactory, create_engine, init_db


T = TypeVar('T')

@dataclass
class Expected(Generic[T]):
    """Wrapper for expected test results to be shared across multiple test cases"""
    value: T

TestCase = tuple
TestGroup = list[Expected | TestCase]
TestCases = TestGroup | TestCase


def _validate_parameters(param_count: int, test_case: TestCase) -> None:
    if param_count > 1:
        if not isinstance(test_case, tuple):
            raise ValueError(f'Expected tuple of {param_count} parameters but got {test_case}')
        elif len(test_case) != param_count:
            raise ValueError(f'Expected {param_count} parameters but got {len(test_case)}: ({test_case})')


def labeled_tests(tests: dict[str, TestCases]):
    def decorator(test_func: Callable):
        parameter_names = inspect.signature(test_func).parameters.keys()
        param_count = len(parameter_names)
        labels = []
        test_data = []

        for group_label, test_cases in tests.items():
            if not isinstance(test_cases, list):
                test_cases = [test_cases]

            expected = None
            try:
                expected_case = next(case for case in test_cases if isinstance(case, Expected))
                expected = expected_case.value
                test_cases.remove(expected_case)
                found_expected = True
            except StopIteration:
                found_expected = False

            for test_case in test_cases if isinstance(test_cases, list) else [test_cases]:
                if found_expected:
                    if isinstance(test_case, tuple):
                        test_case = (*test_case, expected)
                    else:
                        test_case = (test_case, expected)
                _validate_parameters(param_count, test_case)
                labels.append(group_label)
                test_data.append(test_case)
        return pytest.mark.parametrize(', '.join(parameter_names), test_data, ids=labels)(test_func)

    return decorator


class TestDAOFactory(DAOFactory):
    """
    A DAOFactory specifically designed for pytest.
    Includes functionality that can assert what is committed within the DB (through a secondary Session).

    :param debug: If True, uses a test.db file instead of an in-memory SQLite DB. (DB file must be deleted to rerun.
    """
    def __init__(self, debug: bool = False):
        engine = create_engine() if not debug else create_engine('test.db')
        init_db(engine)
        super().__init__(sessionmaker(bind=engine))

    def __enter__(self) -> 'TestDAOFactory':
        super().__enter__()
        return self

    def assert_in_db(self, model: type[DAOModel], *pk, **expected_values: Any) -> None:
        """
        Assert that an object with specific attribute values is present in the DB.
        This checks the committed state of the database, not the session state.

        :param model: The DB table to check
        :param pk: The primary key values of the row
        :param expected_values: The column values to assert
        """
        with self.session_factory() as fresh_session:
            try:
                persisted_copy = DAO(model, fresh_session).get(*pk)
                for key, expected in expected_values.items():
                    actual = getattr(persisted_copy, key)
                    assert actual == expected, f'expected {key} of {persisted_copy} to be {expected} but was {actual}'
            except NotFound as e:
                fail(e.detail)

    def assert_not_in_db(self, model: type[DAOModel], *pk: Any) -> None:
        """
        Assert that the specified object is not present in the DB.
        This checks the committed state of the database, not the session state.

        :param model: The DB table to check
        :param pk: The primary key values of the row
        """
        with self.session_factory() as fresh_session:
            with pytest.raises(NotFound):
                DAO(model, fresh_session).get(*pk)

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)



from typing import Optional, Any
from uuid import UUID

from daomodel import DAOModel
from daomodel.fields import Identifier
from daomodel.testing import labeled_tests, TestDAOFactory
from tests.model_factory import create_test_model
from tests.test_fields import BasicModel
from tests.test_fields__inherited import get_test_cases


class OtherModel(DAOModel, table=True):
    id: Identifier[str]


class UUIDModel(DAOModel, table=True):
    id: Identifier[UUID]


class UUIDReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: UUIDModel


class UUIDOptionalReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[UUIDModel]


@labeled_tests({**get_test_cases('field')})
def test_field(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, value='test')
        entry = dao.find(id=1)
        assert entry.only().value == 'test'


@labeled_tests(get_test_cases('uuid'))
def test_uuid(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        entry = dao.create_with(id=1)

        assert getattr(entry, 'value') is not None
        assert isinstance(getattr(entry, 'value'), UUID)

        entry2 = dao.create_with(id=2)
        assert getattr(entry2, 'value') is not None
        assert isinstance(getattr(entry2, 'value'), UUID)
        assert getattr(entry, 'value') != getattr(entry2, 'value')


@labeled_tests(get_test_cases('dict'))
def test_dict(annotation: Any):
    model_type = create_test_model(annotation)
    sample_dict = {'Hello': 'World!', 'Lorem': ['Ipsum', 'Dolor', 'Sit', 'Amet']}
    with TestDAOFactory() as daos:
        daos[model_type].create_with(id=1, value=sample_dict)
        daos.assert_in_db(model_type, 1, value=sample_dict)


@labeled_tests(get_test_cases('reference'))
def test_reference(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        daos[BasicModel].create(100)
        dao = daos[model_type]
        dao.create_with(id=1, value=100)
        entry = dao.find(id=1)
        assert entry.only().value == 100


def test_reference__uuid(daos: TestDAOFactory):
    uuid_dao = daos[UUIDModel]
    reference_dao = daos[UUIDReferenceModel]

    uuid_model = daos[UUIDModel].create_with()
    reference_dao.create_with(id=1, other_id=uuid_model.id)

    entry = reference_dao.get(1)
    assert entry.other_id is not None
    assert uuid_dao.get(entry.other_id) is not None


def test_reference__uuid__optional(daos: TestDAOFactory):
    daos[UUIDOptionalReferenceModel].create(1)
    daos.assert_in_db(UUIDOptionalReferenceModel, 1, other_id=None)


@labeled_tests(get_test_cases('no_case_str', exclude='unsearchable'))
def test_no_case_str(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, value='TeStVaLuE')

        assert dao.find(value='testvalue').only().id == 1
        assert dao.find(value='TESTVALUE').only().id == 1
        assert dao.find(value='testValue').only().id == 1

        assert dao.find(id=1).only().value == 'TeStVaLuE'



from typing import Any, Iterable, Optional

from pydantic_core import PydanticUndefined
from sqlmodel import SQLModel
from sqlalchemy import Column, Engine, MetaData, Connection, ForeignKey
from str_case_util import Case
from sqlalchemy.ext.declarative import declared_attr

from daomodel.list_util import in_order
from daomodel.metaclass import DAOModelMetaclass
from daomodel.util import reference_of, names_of, retain_in_dict, remove_from_dict
from daomodel.property_filter import PropertyFilter, ALL


ColumnBreadcrumbs = tuple[type['DAOModel'], ..., Column]


class DAOModel(SQLModel, metaclass=DAOModelMetaclass):
    """An SQLModel specifically designed to support a DAO."""

    @declared_attr
    def __tablename__(self) -> str:
        return self.normalized_name()

    @classmethod
    def has_column(cls, column: Column) -> bool:
        """Returns True if the specified Column belongs to this DAOModel."""
        return column.table.name == cls.__tablename__

    @classmethod
    def normalized_name(cls) -> str:
        """A normalized version of this Model name.

        :return: The model name in snake_case form
        """
        return Case.SNAKE_CASE.format(cls.__name__)

    @classmethod
    def doc_name(cls) -> str:
        """A reader-friendly version of this Model name to be used within documentation.

        :return: The model name in Title Case
        """
        return Case.TITLE_CASE.format(cls.__name__)

    @classmethod
    def get_pk(cls) -> Iterable[Column]:
        """Returns the Columns that comprise the Primary Key for this Model.

        :return: A list of primary key columns
        """
        return cls.__table__.primary_key

    @classmethod
    def get_pk_names(cls) -> list[str]:
        """Returns the names of Columns that comprise the Primary Key for this Model.

        :return: A list (of str) of the primary key
        """
        return names_of(cls.get_pk())

    def get_pk_values(self) -> tuple:
        """Returns the values that comprise the Primary Key for this instance of the Model.

        :return: A tuple of primary key values
        """
        return tuple(getattr(self, key) for key in names_of(self.get_pk()))

    def get_pk_dict(self) -> dict[str, Any]:
        """Returns the Primary Key values for this instance of the Model.

        :return: A dict of primary key names/values
        """
        return self.model_dump(include=set(self.get_pk_names()))

    @classmethod
    def get_fks(cls) -> set[Column]:
        """Returns the Columns of other tables that are represented by Foreign Keys for this Model.

        A returned Column could be within this Model in the case of a cyclic relationship.

        :return: An unordered set of columns
        """
        return {fk.column for fk in cls.__table__.foreign_keys}

    @classmethod
    def get_fk_properties(cls) -> set[Column]:
        """Returns the Columns of this Model that represent Foreign Keys.

        :return: An unordered set of foreign key columns
        """
        return {fk.parent for fk in cls.__table__.foreign_keys}

    @classmethod
    def get_references_of(cls, model: type['DAOModel']) -> set[ForeignKey]:
        """Returns the Columns of this Model that represent Foreign Keys of the specified Model.

        :return: An unordered set of foreign key columns
        """
        return {fk for fk in cls.__table__.foreign_keys if model.has_column(fk.column)}

    @classmethod
    def get_properties(cls) -> Iterable[Column]:
        """Returns all the Columns for this Model.

        Column order will match order they are defined in code.
        Inherited properties will be listed first.

        :return: A list of columns
        """
        return cls.__table__.c

    def get_property_names(self, *filters: PropertyFilter) -> list[str]:
        """Returns the names of the specified properties for this record.

        Requested property categories may be refined through filters:

        - `ALL`: All properties
        - `PK`: Primary Key properties
        - `FK`: Foreign Key properties
        - `DEFAULT`: Properties that are equivalent to their default value
        - `NONE`: Properties that do not have a value

        ```python
        # Get all properties (default if no filter specified)
        model.get_property_names()

        # Get all primary key properties
        model.get_property_names(PK)
        ```

        Each filter can be negated by prepending `~` to indicate _NOT_:
        ```python
        # Get all non-null properties
        model.get_property_names(~NONE)

        # Get all properties that are not relationships
        model.get_property_names(~FK)
        ```

        Operators allow for combining filters into expressions:

        - `&` (AND): Properties must match both filters
        - `|` (OR): Properties must match at least one filter
        - `~` (NOT): Properties must NOT match the filter

        ```python
        # Get missing relationships
        model.get_property_names(FK & NONE)

        # Get properties that are a primary or foreign key
        model.get_property_names(PK | FK)

        # Get properties that are their default value or null
        model.get_property_names(DEFAULT | NONE)

        # Get properties that are either not null or are primary key relationships
        model.get_property_names(~NONE | PK & FK)
        ```

        Multiple filter arguments (seperated by commas) are combined with AND:
        ```python
        # Get primary keys that aren't foreign keys
        model.get_property_names(PK, ~FK)  # equivalent to: PK & ~FK

        # Make sure they are also not null
        model.get_property_names(PK, ~FK, ~NONE)  # equivalent to: PK & ~FK & ~NONE
        ```

        Combine several filters to form complex expressions:
        ```python
        # Get properties that are either primary keys that aren't foreign keys or are non-null default values
        model.get_property_names(PK & ~FK | DEFAULT & ~NONE)
        ```

        The filters within an expression are resolved in a specific order:

        1. `~` (NOT): Properties must NOT match the filter
        2. `&` (AND): Properties must match both filters
        3. `|` (OR): Properties must match at least one filter

        Therefore, the following expressions are all equivalent:
        ```python
        model.get_property_names(~NONE | PK & ~FK)
        model.get_property_names(PK & ~FK | ~NONE)
        model.get_property_names(~FK & PK | ~NONE)
        ```

        To change the order of evaluation, use parentheses:
        ```python
        # properties that are either not null or are primary keys but not foreign keys
        model.get_property_names(~FK & PK | ~NONE)

        # properties that are either not null or are not both primary and foreign keys
        model.get_property_names(~(FK & PK) | ~NONE)

        # properties that are not foreign keys and properties that are either primary keys or are non-null
        model.get_property_names(~FK & (PK | ~NONE))
        ```

        :param filters: Property filter expressions using PK, FK, DEFAULT, NONE. Multiple filters are combined with AND.
        :return: A list of property names in the order they are defined within the code
        """
        match len(filters):
            case 0:
                prop_filter = ALL
            case 1:
                prop_filter = filters[0]
            case _:
                prop_filter = filters[0]
                for next_filter in filters[1:]:
                    prop_filter &= next_filter

        result = prop_filter.evaluate(self)
        return in_order(result, names_of(self.get_properties()))

    def get_property_values(self, *filters: PropertyFilter) -> dict[str, Any]:
        """Reads values of the specified properties for this record.

        :param filters: Property filter expressions (see `get_property_names`)
        :return: A dict of property names and their values
        """
        return self.get_values_of(self.get_property_names(*filters))

    @classmethod
    def get_default(cls, column: Column|str):
        """Returns the default value for a given column

        :param column: The Column, or column name
        :return: The default value for the column
        :raises ValueError: if the column has no default value set
        """
        if not isinstance(column, str):
            column = column.name
        field = cls.model_fields.get(column)
        default = field.get_default()
        if default == PydanticUndefined:
            raise ValueError(f'No default is set for {field}.')
        return default

    def get_value_of(self, column: Column|str) -> Any:
        """Shortcut function to return the value for the specified Column.

        :param column: The Column, or column name, to read
        :raises `AttributeError`: if the column is not found.
        """
        if not isinstance(column, str):
            column = column.name
        return getattr(self, column)

    def get_values_of(self, columns: Iterable[Column|str]) -> dict[str, Any]:
        """Reads the values of multiple columns.

        :param columns: The Columns, or their names, to read
        :return: A dict of the column names and their values
        """
        return {column: self.get_value_of(column) for column in columns}

    @classmethod
    def get_searchable_properties(cls) -> Iterable[Column | ColumnBreadcrumbs]:
        """Returns all the Columns for this Model that may be searched using the DAO find function.

        All properties are searchable unless marked with the Unsearchable type annotation.

        To mark a property as unsearchable:
        ```python
        class MyModel(DAOModel, table=True):
            id: Identifier[int]  # Searchable by default
            name: str  # Searchable by default
            internal_notes: Unsearchable[str]  # Not searchable
        ```

        Properties of related models are only searchable if defined within your model's Meta class.
        Please readthedocs for more information.

        :return: A list of searchable columns
        """
        unsearchable = getattr(getattr(cls, '_unsearchable', None), 'default', set())
        searchable = [column for column in cls.get_properties() if column.name not in unsearchable]
        searchable.extend(getattr(getattr(cls, 'Meta', None), 'searchable_relations', set()))
        return searchable

    @classmethod
    def find_searchable_column(cls, prop: str|Column, foreign_tables: list[type['DAOModel']]) -> Column:
        """Returns the specified searchable Column.

        :param prop: str type reference of the Column or the Column itself
        :param foreign_tables: A list of foreign tables to be populated with tables of properties deemed to be foreign
        :return: The searchable Column
        :raises Unsearchable: if the property is not Searchable for this class
        """
        if type(prop) is not str:
            prop = reference_of(prop)
        for column in cls.get_searchable_properties():
            tables = []
            if type(column) is tuple:
                tables = column[:-1]
                column = column[-1]
            if reference_of(column) in [prop, f'{cls.normalized_name()}.{prop}']:
                foreign_tables.extend([t.__table__ for t in tables])
                if column.table is not cls.__table__:
                    foreign_tables.append(column.table)
                return column
        raise UnsearchableError(prop, cls)

    @classmethod
    def pk_values_to_dict(cls, *pk_values: Any) -> dict[str, Any]:
        """Converts the primary key values to a dictionary.

        :param pk_values: The primary key values, in order
        :return: A new dict containing the primary key values
        """
        return dict(zip(cls.get_pk_names(), *pk_values))

    def copy_model(self, source: 'DAOModel', *fields: str) -> None:
        """Copies values from another instance of this Model.

        Unless the fields are specified, all but PK are copied.

        :param source: The model instance from which to copy values
        :param fields: The names of fields to copy
        """
        if fields:
            values = source.model_dump(include=set(fields))
        else:
            values = source.model_dump(exclude=set(source.get_pk_names()))
        self.set_values(**values)

    def set_values(self, ignore_pk: Optional[bool] = False, **values: Any) -> None:
        """Copies property values to this model instance.

        By default, Primary Key values are set if present within the values.

        :param ignore_pk: True if you also wish to not set Primary Key values
        :param values: The dict including values to set
        """
        values = retain_in_dict(values, *names_of(self.get_properties()))
        if ignore_pk:
            values = remove_from_dict(values, *self.get_pk_names())
        for k, v in values.items():
            setattr(self, k, v)

    def __eq__(self, other: 'DAOModel') -> bool:
        """Instances are determined to be equal based on only their primary key."""
        return self.get_pk_values() == other.get_pk_values() if type(self) == type(other) else False

    def __hash__(self) -> int:
        return hash(self.get_pk_values())

    def __str__(self) -> str:
        """
        str representation of this is a str of the primary key.
        A single-column PK results in a simple str value of said column i.e. '1234'
        A multi-column PK results in a str of tuple of PK values i.e. ('Cod', '123 Lake Way')
        """
        pk_values = self.get_pk_values()
        if len(pk_values) == 1:
            pk_values = pk_values[0]
        return str(pk_values)


class UnsearchableError(Exception):
    """Indicates that the Search Query is not allowed for the specified field."""
    def __init__(self, prop: str, model: type(DAOModel)):
        self.detail = f'Cannot search for {prop} of {model.doc_name()}'


def all_models(bind: [Engine|Connection]) -> set[type[DAOModel]]:
    """Discovers all DAOModel types that have been created for the database.

    :param bind: The Engine or Connection for the DB
    :return: A set of applicable DAOModels
    """
    def daomodel_subclasses(cls: type[DAOModel]) -> set[type[DAOModel]]:
        """Returns all defined DAOModels"""
        subclasses = set(cls.__subclasses__())
        for subclass in subclasses.copy():
            subclasses.update(daomodel_subclasses(subclass))
        return subclasses

    metadata = MetaData()
    metadata.reflect(bind=bind)
    db_tables = metadata.tables.keys()
    return {model for model in daomodel_subclasses(DAOModel) if model.__tablename__ in db_tables}



from daomodel import all_models
from daomodel.db import create_engine, init_db
from tests.school_models import *


@pytest.mark.xfail(reason='Fails unless run individually (to better control loaded models)')
def test_all_models():
    engine = create_engine()
    init_db(engine)
    expected = {Person, Book, Hall, Locker, Staff, Student}
    assert all_models(engine) == expected
    connection = engine.connect()
    assert all_models(connection) == expected



from decimal import Decimal
from enum import Enum, auto
from typing import Optional, Any

import pytest

from daomodel import DAOModel
from daomodel.fields import Identifier
from daomodel.model_diff import ModelDiff, Preference, PreferenceRule
from daomodel.testing import labeled_tests, Expected


class LaundryStatus(Enum):
    Public = auto()
    Shared = auto()
    Private = auto()


class Rental(DAOModel, table=True):
    address: Identifier[str]
    apt: Identifier[Optional[str]]
    dwelling_type: str
    sqft: int
    bedrooms: int
    bathrooms: Decimal
    garage_parking: int = 0
    laundry: Optional[LaundryStatus]
    cost: int


one_full = Decimal('1')
one_full_one_half = Decimal('1.1')
two_full = Decimal('2')
two_full_two_half = Decimal('2.2')


dorm = Rental(
    address='123 College Ave',
    apt='12',
    dwelling_type='Dormitory',
    sqft=200,
    bedrooms=1,
    bathrooms=one_full,
    laundry=LaundryStatus.Public,
    cost=0
)
apartment = Rental(
    address='456 City Plaza',
    apt='101',
    dwelling_type='Apartment',
    sqft=850,
    bedrooms=2,
    bathrooms=one_full,
    garage_parking=1,
    laundry=LaundryStatus.Shared,
    cost=1200
)
apartment_two = Rental(
    address='456 City Plaza',
    apt='102',
    dwelling_type='Apartment',
    sqft=850,
    bedrooms=2,
    bathrooms=one_full,
    garage_parking=1,
    laundry=LaundryStatus.Shared,
    cost=1200
)
multi_family = Rental(
    address='789 Suburb St',
    apt='B',
    dwelling_type='Multi-family House',
    sqft=950,
    bedrooms=1,
    bathrooms=one_full_one_half,
    cost=1800
)
town_home = Rental(
    address='321 Maple Dr',
    dwelling_type='Town home',
    sqft=1400,
    bedrooms=3,
    bathrooms=two_full,
    garage_parking=1,
    laundry=LaundryStatus.Private,
    cost=2200
)
single_family = Rental(
    address='987 Country Rd',
    dwelling_type='House',
    sqft=1800,
    bedrooms=4,
    bathrooms=two_full_two_half,
    garage_parking=2,
    laundry=LaundryStatus.Private,
    cost=3000
)


def prefer_better_dwelling_type(values: list[str]) -> Preference:
    """Prefer better dwelling types based on predefined categories."""
    left, right = values
    bad = ['Apartment', 'Multi-family House']
    good = ['Town home', 'House']

    def get_value(dwelling_type: str):
        if dwelling_type in bad:
            return -1
        elif dwelling_type in good:
            return 1
        else:
            return 0

    left_value = get_value(left)
    right_value = get_value(right)

    return (
        Preference.LEFT if left_value > right_value else
        Preference.RIGHT if right_value > left_value else
        Preference.BOTH if left_value > 0 else
        Preference.NEITHER
    )

def prefer_more_private(values: list[LaundryStatus]) -> Preference:
    left, right = values
    left_value = 0 if left is None else left.value
    right_value = 0 if right is None else right.value
    return Preference.LEFT if left_value > right_value else Preference.RIGHT

RENTAL_PREFERENCE_RULES = {
    'address': Preference.NOT_APPLICABLE,
    'apt': Preference.NOT_APPLICABLE,
    'default': max,
    'cost': min,
    'dwelling_type': prefer_better_dwelling_type,
    'laundry': prefer_more_private
}


@labeled_tests({
    'dorm vs apartment':
        (dorm, apartment, {
            'dwelling_type': ('Dormitory', 'Apartment'),
            'sqft': (200, 850),
            'bedrooms': (1, 2),
            'garage_parking': (0, 1),
            'laundry': (LaundryStatus.Public, LaundryStatus.Shared),
            'cost': (0, 1200)
        }),
    'dorm vs multi family':
        (dorm, multi_family, {
            'dwelling_type': ('Dormitory', 'Multi-family House'),
            'sqft': (200, 950),
            'bathrooms': (Decimal('1'), Decimal('1.1')),
            'laundry': (LaundryStatus.Public, None),
            'cost': (0, 1800)
        }),
    'dorm vs town home':
        (dorm, town_home, {
            'dwelling_type': ('Dormitory', 'Town home'),
            'sqft': (200, 1400),
            'bedrooms': (1, 3),
            'bathrooms': (Decimal('1'), Decimal('2')),
            'garage_parking': (0, 1),
            'laundry': (LaundryStatus.Public, LaundryStatus.Private),
            'cost': (0, 2200)
        }),
    'dorm vs single family':
        (dorm, single_family, {
            'dwelling_type': ('Dormitory', 'House'),
            'sqft': (200, 1800),
            'bedrooms': (1, 4),
            'bathrooms': (Decimal('1'), Decimal('2.2')),
            'garage_parking': (0, 2),
            'laundry': (LaundryStatus.Public, LaundryStatus.Private),
            'cost': (0, 3000)
        }),
    'apartment vs multi family':
        (apartment, multi_family, {
            'dwelling_type': ('Apartment', 'Multi-family House'),
            'sqft': (850, 950),
            'bedrooms': (2, 1),
            'bathrooms': (Decimal('1'), Decimal('1.1')),
            'garage_parking': (1, 0),
            'laundry': (LaundryStatus.Shared, None),
            'cost': (1200, 1800)
        }),
    'apartment vs town home':
        (apartment, town_home, {
            'dwelling_type': ('Apartment', 'Town home'),
            'sqft': (850, 1400),
            'bedrooms': (2, 3),
            'bathrooms': (Decimal('1'), Decimal('2')),
            'laundry': (LaundryStatus.Shared, LaundryStatus.Private),
            'cost': (1200, 2200)
        }),
    'apartment vs single family':
        (apartment, single_family, {
            'dwelling_type': ('Apartment', 'House'),
            'sqft': (850, 1800),
            'bedrooms': (2, 4),
            'bathrooms': (Decimal('1'), Decimal('2.2')),
            'garage_parking': (1, 2),
            'laundry': (LaundryStatus.Shared, LaundryStatus.Private),
            'cost': (1200, 3000)
        }),
    'multi family vs town home':
        (multi_family, town_home, {
            'dwelling_type': ('Multi-family House', 'Town home'),
            'sqft': (950, 1400),
            'bedrooms': (1, 3),
            'bathrooms': (Decimal('1.1'), Decimal('2')),
            'garage_parking': (0, 1),
            'laundry': (None, LaundryStatus.Private),
            'cost': (1800, 2200)
        }),
    'multi family vs single family':
        (multi_family, single_family, {
            'dwelling_type': ('Multi-family House', 'House'),
            'sqft': (950, 1800),
            'bedrooms': (1, 4),
            'bathrooms': (Decimal('1.1'), Decimal('2.2')),
            'garage_parking': (0, 2),
            'laundry': (None, LaundryStatus.Private),
            'cost': (1800, 3000)
        }),
    'town home vs single family':
        (town_home, single_family, {
            'dwelling_type': ('Town home', 'House'),
            'sqft': (1400, 1800),
            'bedrooms': (3, 4),
            'bathrooms': (Decimal('2'), Decimal('2.2')),
            'garage_parking': (1, 2),
            'cost': (2200, 3000)
        }),
    'empty diff':
        (apartment, apartment_two, {})
})
def test_model_diff(left: Rental, right: Rental, expected: dict[str, tuple[Any, Any]]):
    assert ModelDiff(left, right) == expected


@labeled_tests({
    'same address':
        (apartment, apartment_two, {
            'apt': ('101', '102')
        }),
    'right None':
        (multi_family, town_home, {
            'address': ('789 Suburb St', '321 Maple Dr'),
            'apt': ('B', None)
        }),
    'left None':
        (single_family, multi_family, {
            'address': ('987 Country Rd', '789 Suburb St'),
            'apt': (None, 'B')
        }),
    'both None':
        (town_home, single_family, {
            'address': ('321 Maple Dr', '987 Country Rd')
        })
})
def test_model_diff__pk(left: Rental, right: Rental, expected: dict[str, tuple[Any, Any]]):
    diff = ModelDiff(left, right, include_pk=True)
    pk_diff = {key: diff[key] for key in ['address', 'apt'] if key in diff}
    assert pk_diff == expected


def test_has_left_value_has_right_value():
    diff = ModelDiff(dorm, apartment)
    assert diff.has_left_value('sqft')
    assert diff.has_right_value('sqft')


def test_get_left_get_right():
    diff = ModelDiff(dorm, apartment)
    assert diff.get_left('sqft') == 200
    assert diff.get_right('sqft') == 850


def test_get_left_get_right__missing():
    diff = ModelDiff(apartment, apartment_two)
    with pytest.raises(KeyError):
        diff.get_left('address')
    with pytest.raises(KeyError):
        diff.get_right('cost')


def test_get_left_get_right__invalid():
    diff = ModelDiff(apartment, apartment_two)
    with pytest.raises(KeyError):
        diff.get_left('APT')
    with pytest.raises(KeyError):
        diff.get_right('invalid')


def test_find_rule():
    assert ModelDiff(apartment, apartment_two, **RENTAL_PREFERENCE_RULES)._find_rule('apt') == Preference.NOT_APPLICABLE


@labeled_tests({
    'Preference': [
        Preference.LEFT,
        Preference.RIGHT,
        Preference.BOTH,
        Preference.NEITHER,
        Preference.NOT_APPLICABLE
    ]
})
def test_execute_rule__preference(rule: PreferenceRule):
    assert ModelDiff.execute_rule(rule, ['left', 'right']) == rule


def named_args_rule(left: str, right: str) -> Preference:
    if left == 'good':
        return Preference.LEFT
    elif right == 'good':
        return Preference.RIGHT
    else:
        return Preference.NEITHER


def var_args_rule(*args: str) -> Preference:
    return named_args_rule(args[0], args[1])


def list_arg_rule(values: list[str]) -> Preference:
    return named_args_rule(values[0], values[1])


@labeled_tests({
    'named args': named_args_rule,
    'var args': var_args_rule,
    'list': list_arg_rule,
    'lambda named args': lambda left, right: (
        Preference.LEFT if left == 'good'
        else Preference.RIGHT if right == 'good'
        else Preference.NEITHER
    ),
    'lambda var args': lambda *args: (
        Preference.LEFT if args[0] == 'good'
        else Preference.RIGHT if args[1] == 'good'
        else Preference.NEITHER
    ),
    'lambda list': lambda values: (
        Preference.LEFT if values[0] == 'good'
        else Preference.RIGHT if values[1] == 'good'
        else Preference.NEITHER
    )
})
def test_execute_rule(rule: PreferenceRule):
    assert ModelDiff.execute_rule(rule, ['bad', 'good']) == Preference.RIGHT


@labeled_tests({
    'number': [-1, 0, 1000, 0.5],
    'str': ['a', 'b', 'c', 'left', 'right'],
    'bool': [True, False],
    'none': [None]
})
def test_execute_rule__return_non_preference(value: Any | None):
    assert ModelDiff.execute_rule(lambda _: value, ['left', 'right']) == value


@labeled_tests({
    'already preference': [
        (Preference.LEFT, 'dwelling_type', Preference.LEFT),
        (Preference.RIGHT, 'dwelling_type', Preference.RIGHT),
        (Preference.NEITHER, 'dwelling_type', Preference.NEITHER),
        (Preference.BOTH, 'dwelling_type', Preference.BOTH),
        (Preference.NOT_APPLICABLE, 'dwelling_type', Preference.NOT_APPLICABLE)
    ],
    'equals left value': [Expected(Preference.LEFT),
        ('Dormitory', 'dwelling_type'),
        (200, 'sqft'),
        (1, 'bedrooms'),
        (0, 'garage_parking'),
        (LaundryStatus.Public, 'laundry')
    ],
    'equals right value': [Expected(Preference.RIGHT),
        ('House', 'dwelling_type'),
        (1800, 'sqft'),
        (4, 'bedrooms'),
        (2, 'garage_parking'),
        (LaundryStatus.Private, 'laundry')
    ],
    'equals neither value': [Expected(Preference.NEITHER),
        ('unknown', 'dwelling_type'),
        (1000, 'sqft'),
        (2, 'bedrooms'),
        (3, 'garage_parking'),
        (LaundryStatus.Shared, 'laundry'),
        (None, 'dwelling_type')
    ]
})
def test_map_resolution_to_preference(resolution: Preference | Any | None, field: str, expected: Preference):
    assert ModelDiff(dorm, single_family).map_resolution_to_preference(resolution, field) == expected


def test_get_preferred__not_applicable():
    diff = ModelDiff(multi_family, single_family, include_pk=True, **RENTAL_PREFERENCE_RULES)
    assert diff.get_preferred('address') == Preference.NOT_APPLICABLE
    assert diff.get_preferred('apt') == Preference.NOT_APPLICABLE


def test_get_preferred__not_implemented():
    with pytest.raises(NotImplementedError):
        ModelDiff(apartment, multi_family).get_preferred('cost')


def test_get_preferred__missing():
    with pytest.raises(KeyError):
        ModelDiff(single_family, town_home).get_preferred('laundry')


class RentalDiff(ModelDiff[Rental]):
    def __init__(self, left: Rental, right: Rental):
        super().__init__(left, right, **RENTAL_PREFERENCE_RULES)

@labeled_tests({
    'dorm vs apartment': [
        (dorm, apartment, 'dwelling_type', Preference.LEFT),
        (dorm, apartment, 'sqft', Preference.RIGHT),
        (dorm, apartment, 'bedrooms', Preference.RIGHT),
        (dorm, apartment, 'garage_parking', Preference.RIGHT),
        (dorm, apartment, 'laundry', Preference.RIGHT),
        (dorm, apartment, 'cost', Preference.LEFT)
    ],
    'dorm vs multi family': [
        (dorm, multi_family, 'dwelling_type', Preference.LEFT),
        (dorm, multi_family, 'sqft', Preference.RIGHT),
        (dorm, multi_family, 'bathrooms', Preference.RIGHT),
        (dorm, multi_family, 'laundry', Preference.LEFT),
        (dorm, multi_family, 'cost', Preference.LEFT)
    ],
    'dorm vs town home': [
        (dorm, town_home, 'dwelling_type', Preference.RIGHT),
        (dorm, town_home, 'sqft', Preference.RIGHT),
        (dorm, town_home, 'bedrooms', Preference.RIGHT),
        (dorm, town_home, 'bathrooms', Preference.RIGHT),
        (dorm, town_home, 'garage_parking', Preference.RIGHT),
        (dorm, town_home, 'laundry', Preference.RIGHT),
        (dorm, town_home, 'cost', Preference.LEFT)
    ],
    'dorm vs single family': [
        (dorm, single_family, 'dwelling_type', Preference.RIGHT),
        (dorm, single_family, 'sqft', Preference.RIGHT),
        (dorm, single_family, 'bedrooms', Preference.RIGHT),
        (dorm, single_family, 'bathrooms', Preference.RIGHT),
        (dorm, single_family, 'garage_parking', Preference.RIGHT),
        (dorm, single_family, 'laundry', Preference.RIGHT),
        (dorm, single_family, 'cost', Preference.LEFT)
    ],
    'apartment vs multi family': [
        (apartment, multi_family, 'dwelling_type', Preference.NEITHER),
        (apartment, multi_family, 'sqft', Preference.RIGHT),
        (apartment, multi_family, 'bedrooms', Preference.LEFT),
        (apartment, multi_family, 'bathrooms', Preference.RIGHT),
        (apartment, multi_family, 'garage_parking', Preference.LEFT),
        (apartment, multi_family, 'laundry', Preference.LEFT),
        (apartment, multi_family, 'cost', Preference.LEFT)
    ],
    'apartment vs town home': [
        (apartment, town_home, 'dwelling_type', Preference.RIGHT),
        (apartment, town_home, 'sqft', Preference.RIGHT),
        (apartment, town_home, 'bedrooms', Preference.RIGHT),
        (apartment, town_home, 'bathrooms', Preference.RIGHT),
        (apartment, town_home, 'laundry', Preference.RIGHT),
        (apartment, town_home, 'cost', Preference.LEFT)
    ],
    'apartment vs single family': [
        (apartment, single_family, 'dwelling_type', Preference.RIGHT),
        (apartment, single_family, 'sqft', Preference.RIGHT),
        (apartment, single_family, 'bedrooms', Preference.RIGHT),
        (apartment, single_family, 'bathrooms', Preference.RIGHT),
        (apartment, single_family, 'garage_parking', Preference.RIGHT),
        (apartment, single_family, 'laundry', Preference.RIGHT),
        (apartment, single_family, 'cost', Preference.LEFT)
    ],
    'multi family vs town home': [
        (multi_family, town_home, 'dwelling_type', Preference.RIGHT),
        (multi_family, town_home, 'sqft', Preference.RIGHT),
        (multi_family, town_home, 'bedrooms', Preference.RIGHT),
        (multi_family, town_home, 'bathrooms', Preference.RIGHT),
        (multi_family, town_home, 'garage_parking', Preference.RIGHT),
        (multi_family, town_home, 'laundry', Preference.RIGHT),
        (multi_family, town_home, 'cost', Preference.LEFT)
    ],
    'multi family vs single family': [
        (multi_family, single_family, 'dwelling_type', Preference.RIGHT),
        (multi_family, single_family, 'sqft', Preference.RIGHT),
        (multi_family, single_family, 'bedrooms', Preference.RIGHT),
        (multi_family, single_family, 'bathrooms', Preference.RIGHT),
        (multi_family, single_family, 'garage_parking', Preference.RIGHT),
        (multi_family, single_family, 'laundry', Preference.RIGHT),
        (multi_family, single_family, 'cost', Preference.LEFT)
    ],
    'town home vs single family': [
        (town_home, single_family, 'dwelling_type', Preference.BOTH),
        (town_home, single_family, 'sqft', Preference.RIGHT),
        (town_home, single_family, 'bedrooms', Preference.RIGHT),
        (town_home, single_family, 'bathrooms', Preference.RIGHT),
        (town_home, single_family, 'garage_parking', Preference.RIGHT),
        (town_home, single_family, 'cost', Preference.LEFT)
    ]
})
def test_get_preferred__custom_diff(left: Rental, right: Rental, field: str, expected: Preference):
    assert RentalDiff(left, right).get_preferred(field) == expected



from typing import Any

import pytest
from sqlalchemy.exc import IntegrityError

from daomodel import names_of
from daomodel.testing import labeled_tests, TestDAOFactory
from tests.model_factory import create_test_model
from tests.test_fields import BasicModel
from tests.test_fields__inherited import get_test_cases


@labeled_tests(get_test_cases(exclude=('optional', 'uuid')))
def test_required(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        entry = dao.create_with(id=1, insert=False)
        with pytest.raises(IntegrityError):
            dao.insert(entry)


@labeled_tests(get_test_cases('uuid', exclude='optional'))
def test_required__uuid(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1)
        entry = dao.find(id=1).only()
        # Since UUID has a default value, we test that it is Required by attempting to set it to None
        entry.value = None
        with pytest.raises(IntegrityError):
            dao.upsert(entry)


@labeled_tests(get_test_cases('optional', exclude='uuid'))
def test_optional(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1)
        entry = dao.find(id=1).only()
        assert entry.value is None


@labeled_tests(get_test_cases('optional', 'uuid'))
def test_optional__uuid(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1)
        entry = dao.find(id=1).only()
        assert entry.value is not None
        # Since UUID has a default value, we test that it is Optional by setting it to None
        entry.value = None
        dao.upsert(entry)
        entry = dao.find(id=1).only()
        assert entry.value is None


def test_identifier():
    assert BasicModel.get_pk_names() == ['id']
    assert create_test_model(str).get_pk_names() == ['id']


@labeled_tests(get_test_cases('identifier'))
def test_identifier__continued(annotation: Any):
    assert create_test_model(annotation, field_name='id2').get_pk_names() == ['id', 'id2']


@labeled_tests(get_test_cases(exclude='unsearchable'))
def test_searchable(annotation: Any):
    assert names_of(create_test_model(annotation).get_searchable_properties()) == ['id', 'value']


@labeled_tests(get_test_cases('unsearchable'))
def test_unsearchable(annotation: Any):
    assert names_of(create_test_model(annotation).get_searchable_properties()) == ['id']



from enum import Enum
from typing import Any, Optional, Callable, TypeVar

from daomodel import DAOModel
from daomodel.property_filter import PK


class Preference(Enum):
    """Represents the preference of values in a model comparison.

    The available options are:
    - LEFT: Indicates preference for the 'left' side value.
    - RIGHT: Indicates preference for the 'right' side value.
    - NEITHER: Indicates that neither value is preferred.
    - BOTH: Indicates that both values are equally preferable.
    - NOT_APPLICABLE: Indicates that a preference does not apply in the given context i.e. the name field of two customers.
    """
    NOT_APPLICABLE = -1
    NEITHER = 0
    LEFT = 1
    RIGHT = 2
    BOTH = 3


PreferenceRuleFunction = Callable[[list], Preference] | Callable[[Any, Any, ...], Preference]
PreferenceRule = PreferenceRuleFunction | Preference


T = TypeVar('T', bound=DAOModel)


class ModelDiff(dict[str, tuple[Any, Any]]):
    """A dictionary wrapper of differing property values between two DAOModel instances.

    While designed to compare like models, it should work between different model types. Though that is untested.

    The keys in the dictionary represent the property names, and the value for the key contains each object value.
    Properties that share the same value across the objects are excluded.

    :param left: The left model instance used for comparison
    :param right: The right model instance used for comparison
    :param include_pk: True to include primary key values in the diff (False by default)
    :param preference_rules: Dictionary of field names to rules that define value preferences (see `execute_rule`)
    """
    def __init__(self,
                 left: T,
                 right: T,
                 include_pk: Optional[bool] = False,
                 **preference_rules: PreferenceRule):
        super().__init__()
        self.left: T = left
        self.right: T = right
        self.preference_rules = preference_rules

        filter_expr = [] if include_pk else [~PK]
        left_values = left.get_property_values(*filter_expr)
        right_values = right.get_property_values(*filter_expr)

        for k, v in left_values.items():
            if right_values[k] != v:
                self[k] = (v, right_values[k])

    def has_left_value(self, field: str) -> bool:
        """Returns True if the left value for the specified field exists"""
        return self.get_right(field) is not None

    def get_left(self, field: str) -> Any:
        """Fetches the value of the left object.

        :param field: The name of the field to fetch
        :return: The left value for the specified field
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        if field not in self:
            raise KeyError(f'{field} not found in diff.')
        return self.get(field)[0]

    def has_right_value(self, field: str) -> bool:
        """Returns True if the right value for the specified field exists"""
        return self.get_right(field) is not None

    def get_right(self, field: str) -> Any:
        """Fetches the value of the right object.

        :param field: The name of the field to fetch
        :return: The right value for the specified field
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        if field not in self:
            raise KeyError(f'{field} not found in diff.')
        return self.get(field)[1]

    def all_values(self, field: str) -> list[Any]:
        """Returns a list containing the values for the specified field, ordered from left to right."""
        return [self.get_left(field), self.get_right(field)]

    def get_preferred(self, field: str) -> Preference:
        """Determines which of the differing values is preferred.

        :param field: The name of the field
        :return: The Preference between the possible values
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        :raises NotImplementedError: if there is no applicable preference rule provided for the field
        """
        if field not in self:
            raise KeyError(f'{field} not found in diff.')

        rule = self._find_rule(field)
        resolution = self.execute_rule(rule, self.all_values(field))
        preference = self.map_resolution_to_preference(resolution, field)

        return preference

    def _find_rule(self, field: str) -> Callable:
        """Finds the rule for determining the preferred value for the specified field."""
        def raise_error(*_) -> Any:
            raise NotImplementedError(f'No rule is defined to determine preference for {field}')
        return self.preference_rules.get(field, self.preference_rules.get('default', raise_error))

    @staticmethod
    def execute_rule(rule: PreferenceRule, values: list) -> Preference | Optional[Any]:
        """Conducts the rule for the specified field to determine the preferred value.

        A PreferenceRule may be a function or a static Preference i.e. LEFT, RIGHT, BOTH, NEITHER, or NOT_APPLICABLE.
        In the case of a function, it must accept multiple arguments, in the format of multiple args, *args, or a list.
        The return value of the function must be a Preference, one of the field values, or None.

        For example:
            def prefer_longer_name(values: list[str]) -> str:
                return max(values, key=len)

            def prefer_true(left: bool, right: bool) -> Preference:
                if left:
                    return Preference.LEFT
                elif right:
                    return Preference.RIGHT
                else
                    return Preference.NEITHER

        :param rule: The rule to execute
        :param values: The field values to apply the rule against
        :return: The resolved Preference or value based on the rule
        """
        if callable(rule):
            try:
                return rule(*values)
            except TypeError:
                return rule(values)
        return rule

    def map_resolution_to_preference(self, resolution: Preference | Optional[Any], field: str) -> Preference:
        """Converts a given resolution to its corresponding Preference

        :param resolution: The resolved value of Preference for the field
        :param field: The name of the field the resolution applies to
        :return: The appropriate Preference for the resolution
        """
        if isinstance(resolution, Preference):
            return resolution
        elif resolution == self.get_left(field):
            return Preference.LEFT
        elif resolution == self.get_right(field):
            return Preference.RIGHT
        else:
            return Preference.NEITHER



from typing import Any, Iterable, Optional

import pytest

from daomodel.list_util import (
    most_frequent, ensure_iter, dedupe, in_order, strip_whitespace, exclude_falsy, first_str_with, first, last, longest
)
from daomodel.testing import labeled_tests


@labeled_tests({
    'already iterable': [
        ([1, 2, 3], [1, 2, 3]),
        ({1, 2, 3}, {1, 2, 3}),
        ((1, 2, 3), (1, 2, 3))
    ],
    'empty': [
        ([], []),
        ({}, {}),
        ((), ()),
    ],
    'single item': [
        (1, [1]),
        (None, [None]),
        ('element', ['element']),
    ]
})
def test_ensure_iter(elements: Any, expected: Iterable[Any]):
    assert ensure_iter(elements) == expected


@labeled_tests({
    'empty': [
        (set(), [], []),
        (set(), [1], []),
        ({1}, [], []),
    ],
    'single item': [
        ({1}, [1], [1]),
        ({'a'}, ['a'], ['a'])
    ],
    'multiple items': [
        ({1, 2, 3}, [1, 2, 3], [1, 2, 3]),
        ({1, 2, 3}, [3, 2, 1], [3, 2, 1]),
        ({'a', 'b', 'c'}, ['a', 'b', 'c'], ['a', 'b', 'c']),
        ({'a', 'b', 'c'}, ['c', 'b', 'a'], ['c', 'b', 'a'])
    ],
    'repeated items': [
        ([1, 1, 2, 2, 3, 3], [1, 2, 3], [1, 2, 3]),
        ([1, 1, 2, 2, 3, 3], [3, 2, 1], [3, 2, 1]),
        (['a', 'b', 'c', 'b', 'a'], ['a', 'b', 'c'], ['a', 'b', 'c']),
        (['a', 'b', 'c', 'b', 'a'], ['c', 'b', 'a'], ['c', 'b', 'a']),
        ([True, True, False, True, False, False], [False, True], [False, True])
    ],
    'items missing from order': [
        ({1, 2, 3, 4}, [1, 2, 3], [1, 2, 3]),
        ({1, 2, 3, 4}, [3, 2, 1], [3, 2, 1])
    ],
    'extraneous items in order': [
        ({1, 3}, [1, 2, 3], [1, 3]),
        ({1, 3}, [3, 2, 1], [3, 1])
    ],
    'mixed scenarios': [
        ({4, 1, 1, 3, 4}, [1, 2, 3], [1, 3]),
        ({1, 3, 1, 4}, [3, 2, 1], [3, 1])
    ]
})
def test_in_order(items: Iterable, order: list, expected: list):
    assert in_order(items, order) == expected


@pytest.mark.parametrize('elements, expected', [
    ([1, 2, 3], [1, 2, 3]),
    ([1, 1, 2, 2, 3, 3, 3], [1, 2, 3]),
    (['one', 'two', 'two', 'three'], ['one', 'two', 'three']),
    (['one', 1, 'one', 'two', 2, 'three', 2, 'three', 2], ['one', 1, 'two', 2, 'three']),
    ([], []),
])
def test_dedupe(elements: list, expected: list):
    assert dedupe(elements) == expected


def test_dedupe__keep_last():
    elements = ['one', 1, 'one', 'two', 2, 'three', 2, 'three', 2]
    expected = [1, 'one', 'two', 'three', 2]
    assert dedupe(elements, keep_last=True) == expected


@pytest.mark.parametrize('elements, expected', [
    (['  test  ', ' example ', 'no trim'], ['test', 'example', 'no trim']),
    (['  leading', 'trailing  ', '  both  '], ['leading', 'trailing', 'both']),
    (['', '   ', '\t\n'], ['', '', '']),
    ([], []),
])
def test_strip_whitespace(elements: list, expected: list):
    assert strip_whitespace(elements) == expected


@pytest.mark.parametrize('elements, expected', [
    ([1, 2, 2, 3], 2),
    ([1, 1, 2, 2, 3], 1),
    (['a', 'b', 'b', 'c', 'c', 'c'], 'c'),
    ([True, True, False, True], True),
    ([1], 1),
    ([None, 'not none', None, 'not none', None], None)
])
def test_most_frequent(elements: list, expected: Any):
    assert most_frequent(elements) == expected


@pytest.mark.parametrize('elements, expected', [
    (['apple', 'banana', 'cherry'], 'banana'),
    (['a', 'ab', 'abc', 'abcd'], 'abcd'),
    (['same', 'size', 'test'], 'same'),
    ([], None),
    (['one'], 'one'),
    (['long', 'longer', 'longest', 'tiny'], 'longest')
])
def test_longest(elements: list, expected: Any):
    assert longest(elements) == expected


@labeled_tests({
    'all truthy': [
        ([1, 'test', True], [1, 'test', True]),
        ([True, True], [True, True]),
        (['Text'], ['Text'])
    ],
    'all falsy': [
        ([None, False, 0, ''], []),
        ([False, False], []),
        ([0], []),
        ([None], []),
    ],
    'mixed':
        ([0, False, None, '', 1, 'text', True], [1, 'text', True]),
    'empty':
        ([], [])
})
def test_exclude_falsy(elements: list, expected: list):
    assert exclude_falsy(elements) == expected


@labeled_tests({
    'complete match': [
        ('sample', ['sample', 'test', 'content'], 'sample'),
        ('test', ['sample', 'test', 'content'], 'test'),
        ('content', ['sample', 'test', 'content'], 'content')
    ],
    'word match': [
        ('sample', ['sample str', 'test str', 'str content'], 'sample str'),
        ('test', ['sample str', 'test str', 'str content'], 'test str'),
        ('content', ['sample str', 'test str', 'str content'], 'str content'),
    ],
    'partial match': [
        ('s', ['samples', 'testing', 'content'], 'samples'),
        ('test', ['samples', 'testing', 'content'], 'testing'),
        ('tent', ['samples', 'testing', 'content'], 'content'),
    ],
    'multiple matches': [
        ('e', ['sample', 'test', 'content'], 'sample'),
        ('', ['sample', 'test', 'content'], 'sample'),
        ('te', ['sample', 'test', 'content'], 'test'),
        ('content', ['sample content', 'test', 'content'], 'sample content')
    ],
    'no match': [
        ('missing', ['sample', 'test', 'content'], None),
        ('test', [], None)
    ]
})
def test_first_str_with(substring: str, strings: list[str], expected: Optional[str]):
    assert first_str_with(substring, strings) == expected


@labeled_tests({
    'single value': [
        ([1], 1, 1),
        ([True], True, True),
        (['text'], 'text', 'text')
    ],
    'one truthy value':
        ([0, False, 1, None, ''], 1, 1),
    'multiple truthy value': [
        ([1, 2, 3], 1, 3),
        ([True, 23, 'text'], True, 'text')
    ],
    'mixed': [
        ([0, 1, 2, 3, 0], 1, 3),
        ([False, 0, 6, 'text', True, ''], 6, True)
    ],
    'all falsy':
        ([None, False, 0, ''], None, None),
    'empty':
        ([], None, None)
})
def test_first_last(elements: list, expected_first: Any, expected_last: Any):
    assert first(elements) == expected_first
    assert last(elements) == expected_last



from typing import Optional

from sqlmodel import Field, Relationship

from daomodel.backwards_compatibility import DAOModel
from daomodel.testing import TestDAOFactory


class ParentModel(DAOModel, table=True):
    id: int = Field(primary_key=True)
    children: list['ChildModel'] = Relationship(back_populates='parent')
    _private: list[str]


class ChildModel(DAOModel, table=True):
    id: int = Field(primary_key=True)
    sibling: Optional[int] = Field(foreign_key='childmodel.id', ondelete='SET NULL')
    parent_id: Optional[int] = Field(foreign_key='parentmodel.id')
    parent: Optional['ParentModel'] = Relationship(back_populates='children')


def test_table_naming() -> None:
    assert ParentModel.__table__.name == 'parentmodel'
    assert ChildModel.__table__.name == 'childmodel'


def test_relationship_behavior(daos: TestDAOFactory) -> None:
    parent_dao = daos[ParentModel]
    child_dao = daos[ChildModel]
    parent = parent_dao.create(0)
    child = child_dao.create_with(id=1, parent=parent.id)
    child_dao.create_with(id=2, sibling=1)

    daos.assert_in_db(ChildModel, 2, sibling=1)
    child_dao.remove(child)
    daos.assert_in_db(ChildModel, 2, sibling=None)



import pytest
from sqlalchemy.testing.schema import Column

from daomodel import reference_of
from daomodel.util import names_of, values_from_dict, retain_in_dict, remove_from_dict, MissingInput
from tests.school_models import Person, Book


@pytest.mark.parametrize('column, expected', [
    (Person.name, 'person.name'),
    (Person.ssn, 'person.ssn'),
    (Book.owner, 'book.owner')
])
def test_reference_of(column: Column, expected: str):
    assert reference_of(column) == expected


@pytest.mark.parametrize('columns, expected', [
    ([], []),
    ([Column('one')], ['one']),
    ([Column('one'), Column('two'), Column('three')], ['one', 'two', 'three'])
])
def test_names_of(columns: list[Column], expected:  list[str]):
    assert names_of(columns) == expected


@pytest.mark.parametrize('keys, expected', [
    ((), ()),
    (('b',), (2,)),
    (('a', 'c'), (1, 3)),
    (('b', 'c', 'a'), (2, 3, 1))
])
def test_values_from_dict(keys: tuple[str, ...], expected: tuple):
    assert values_from_dict(*keys, a=1, b=2, c=3) == expected


def test_values_from_dict__missing():
    with pytest.raises(MissingInput):
        values_from_dict('a', 'b', 'd', a=1, b=2, c=3)


@pytest.mark.parametrize('keys, expected', [
    ((), {}),
    (('b',), {'b': 2}),
    (('a', 'c'), {'a':1, 'c':3}),
    (('b', 'c', 'a'), {'a':1, 'b':2, 'c':3})
])
def test_retain_in_dict(keys: tuple[str, ...], expected: tuple):
    assert retain_in_dict({'a': 1, 'b': 2, 'c': 3}, *keys) == expected


@pytest.mark.parametrize('keys, expected', [
    ((), {'a':1, 'b':2, 'c':3}),
    (('b',), {'a':1, 'c':3}),
    (('a', 'c'), {'b': 2}),
    (('b', 'c', 'a'), {})
])
def test_remove_from_dict(keys: tuple[str, ...], expected: tuple):
    assert remove_from_dict({'a': 1, 'b': 2, 'c': 3}, *keys) == expected



from typing import Iterable, Any

from sqlalchemy import Column


class MissingInput(Exception):
    """Indicates that required information was not provided."""
    def __init__(self, detail: str):
        self.detail = detail


class InvalidArgumentCount(Exception):
    """Indicates that an incorrect number of arguments was provided."""
    def __init__(self, expected: int, got: int, context: str = None):
        self.detail = f'Expected {expected} values, got {got}'
        if context:
            self.detail += f' for {context}'


class UnsupportedFeatureError(Exception):
    """Indicates that a feature is not yet supported.

    If you think the feature should be implemented, please open an issue on GitHub.
    """
    def __init__(self, detail: str):
        self.detail = detail + (' Note: This functionality is not yet supported. '
                                'Please submit a request through GitHub if you would like it implemented.')


def reference_of(column: Column) -> str:
    """
    Prepares a str reference of a column.

    :param column: The column to convert
    :return: The 'table.column' notation of the Column
    """
    return f'{column.table.name}.{column.name}'


def names_of(properties: Iterable[Column]) -> list[str]:
    """
    Reduces Columns to just their names.

    :param properties: A group of Columns
    :return: A list of names matching the order of the Columns provided
    """
    return [p.name for p in properties]


def values_from_dict(*keys: Any, **values: Any) -> tuple:
    """Pulls specific values from a dictionary.

    :param keys: The keys to read from the dict
    :param values: The dictionary containing the values
    :return: A tuple of values read from the dict, in the same order as keys
    """
    result = []
    for key in keys:
        if key in values:
            result.append(values[key])
        else:
            raise MissingInput(f'Requested key {key} not found in dictionary')
    return tuple(result)


def retain_in_dict(d: dict[Any, Any], *keys: Any) -> dict[Any, Any]:
    """Filters a dictionary to specified keys.

    The source dict remains unmodified.

    :param d: The dictionary to filter down
    :param keys: The target keys for the new dict
    :return: The reduced values as a new dict
    """
    return {key: d[key] for key in keys if key in d}


def remove_from_dict(d: dict[Any, Any], *keys: Any) -> dict[Any, Any]:
    """Removes specified key/value pairs from a dictionary.

    The source dict remains unmodified.

    :param d: The dictionary to adjust
    :param keys: The keys to remove from the dict
    :return: The modified values as a new dict
    """
    return {k: v for k, v in d.items() if k not in keys}


def next_id() -> None:
    """Indicates to the model that an id should be auto-incremented"""
    return None



from typing import Dict, Any, Tuple, Type, get_origin, get_args, Union, Optional, List, ForwardRef
import inspect
import uuid
from sqlmodel.main import SQLModelMetaclass, Field, FieldInfo, RelationshipInfo, Relationship
from sqlalchemy import ForeignKey, JSON, String

from daomodel.util import reference_of, UnsupportedFeatureError
from daomodel.fields import no_case_str, Identifier, Unsearchable, Protected, ReferenceTo


class Annotation:
    """A utility class to help manage a type-annotated field."""
    def __init__(self, field_name: str, field_type: type[Any]):
        self.name = field_name

        self.modifiers = set()
        for modifier in [Unsearchable, Identifier, Protected]:
            if get_origin(field_type) is modifier:
                self.modifiers.add(modifier)
                field_type = get_args(field_type)[0]
        if get_origin(field_type) is Union:
            args = get_args(field_type)
            if len(args) == 2 and args[1] is type(None):
                self.modifiers.add(Optional)
                field_type = args[0]

        self.type = field_type
        self.args = {}

    def is_private(self) -> bool:
        """Check whether the annotation is for a private field."""
        return self.name.startswith('_')

    def has_modifier(self, modifier: Any) -> bool:
        """Check whether the annotation has a specified modifier.

        :param modifier: The modifier to check for, valid modifiers are Unsearchable, Identifier, Protected, Optional
        :return: True if the annotation has the modifier
        """
        return modifier in self.modifiers

    def is_relationship(self) -> bool:
        """Check whether the annotation is a defined relationship to another model"""
        origin = get_origin(self.type)
        if origin not in (list, List):
            return False
        args = get_args(self.type)
        return args and len(args) == 1 and (isinstance(args[0], (str, ForwardRef)) or inspect.isclass(args[0]))

    def is_dao_model(self) -> bool:
        """Check whether the annotation is a DAOModel."""
        return inspect.isclass(self.type) and 'DAOModel' in (base.__name__ for base in inspect.getmro(self.type))

    def __getitem__(self, key: str) -> Any:
        return self.args.get(key)

    def __setitem__(self, key: str, value: Any) -> None:
        self.args[key] = value


class ClassDictHelper:
    """A utility class to help manage class dictionary and annotations in metaclasses."""
    def __init__(self, class_dict: dict[str, Any]):
        self.class_dict = class_dict

    @property
    def annotations(self) -> dict[str, Any]:
        return self.class_dict.get('__annotations__', {})

    def set_annotation(self, field: Annotation) -> None:
        """Set an annotation for a field, automatically handling optional types if nullable is True."""
        self.annotations[field.name] = Union[field.type, None] if field['nullable'] else field.type

    @property
    def fields(self) -> list[Annotation]:
        fields = [Annotation(field_name, field_type) for field_name, field_type in self.annotations.items()]
        return [field for field in fields if not field.is_private() and not self.is_relationship(field)]

    def is_relationship(self, field: Annotation) -> bool:
        return field in self and isinstance(self[field], RelationshipInfo)

    def is_reference(self, field: Annotation) -> bool:
        return field in self and isinstance(self[field], ReferenceTo)

    def add_unsearchable(self, field: Annotation) -> None:
        """Mark a field as unsearchable within in the class dictionary."""
        self.class_dict.setdefault('_unsearchable', set()).add(field.name)

    def __getitem__(self, field: Annotation) -> Any:
        return self.class_dict.get(field.name)

    def __setitem__(self, field: Annotation, value: Any) -> None:
        self.class_dict[field.name] = value

    def __contains__(self, field: Annotation) -> bool:
        return field.name in self.class_dict


class DAOModelMetaclass(SQLModelMetaclass):
    """A metaclass for DAOModel that adds support for modifiers and special typing within annotations."""
    def __new__(
            cls,
            name: str,
            bases: Tuple[Type[Any], ...],
            class_dict: Dict[str, Any],
            **kwargs: Any,
    ) -> Any:
        model = ClassDictHelper(class_dict)

        for field in model.fields:
            if field.is_relationship():
                model[field] = Relationship()
                continue
            cls._process_field_modifiers(field, model)
            cls._process_field_type(field, model)
            model.set_annotation(field)
            cls._process_existing_field(field, model)

        return super().__new__(cls, name, bases, class_dict, **kwargs)

    @classmethod
    def _process_field_modifiers(cls, field: Annotation, model: ClassDictHelper) -> None:
        """Process field modifiers like Unsearchable, Identifier, and Optional."""
        if field.has_modifier(Unsearchable):
            model.add_unsearchable(field)
        if field.has_modifier(Identifier):
            field['primary_key'] = True
        field['nullable'] = field.has_modifier(Optional)

    @classmethod
    def _process_field_type(cls, field: Annotation, model: ClassDictHelper) -> None:
        """Process field type-specific settings."""
        if field.type is uuid.UUID:
            field['default_factory'] = uuid.uuid4
        elif field.type is dict:
            field['sa_type'] = JSON
        elif field.type is no_case_str:
            field.type = str
            field['sa_type'] = String(collation='NOCASE')
        elif model.is_reference(field) or field.is_dao_model():
            cls._process_reference_field(field, model)

    @classmethod
    def _process_reference_field(cls, field: Annotation, model: ClassDictHelper) -> None:
        """Process fields that reference other models, AKA foreign key fields."""
        if field.is_dao_model():
            cls._process_dao_model_reference(field)
        else:
            field['foreign_key'] = getattr(model[field], 'foreign_key')

        field['ondelete'] = cls._determine_ondelete_behavior(field, model)

        field['sa_column_args'] = [
            ForeignKey(
                field['foreign_key'],
                onupdate='CASCADE',
                ondelete=field['ondelete']
            )
        ]

    @classmethod
    def _process_dao_model_reference(cls, field: Annotation) -> None:
        """Process a field that directly references a DAOModel."""
        first_pk = next(iter(field.type.get_pk()))
        if len(field.type.get_pk()) != 1:
            raise UnsupportedFeatureError(
                f'Cannot auto map to composite key of {field.type.__name__}. Use '
                f'Reference(str) instead. i.e. field: int = Reference("{first_pk}")'
            )
        field.type = field.type.__annotations__[first_pk.name]
        field['foreign_key'] = reference_of(first_pk)

    @classmethod
    def _determine_ondelete_behavior(cls, field: Annotation, model: ClassDictHelper) -> str:
        """Determine the appropriate ondelete behavior for a foreign key."""
        existing_value = model[field] if field in model else None
        explicitly_set_value = getattr(existing_value, 'ondelete', None)

        return (
            explicitly_set_value if type(explicitly_set_value) is str else
            'RESTRICT' if field.has_modifier(Protected) else
            'SET NULL' if field['nullable'] else
            'CASCADE'
        )

    @classmethod
    def _process_existing_field(cls, field: Annotation, model: ClassDictHelper) -> None:
        """Process existing field values in the class dictionary."""
        if field in model:
            existing_field = model[field]
            if isinstance(existing_field, FieldInfo):
                for key, value in field.args.items():
                    setattr(existing_field, key, value)
                return
            else:
                field['default'] = existing_field
        model[field] = Field(**field.args)



from typing import Optional, Any

import pytest

from daomodel import DAOModel
from daomodel.fields import Identifier
from daomodel.property_filter import PropertyFilter, BasicPropertyFilter, AndFilter, ALL, PK, FK, DEFAULT, NONE
from daomodel.testing import labeled_tests


class ForeignModel(DAOModel, table=True):
    pk: Identifier[int]


class PropertyModel(DAOModel, table=True):
    pk_value: Identifier[int]
    pk_default_none: Identifier[Optional[int]]
    pk_default: Identifier[int] = 0
    pk_override: Identifier[int] = 0
    pk_default_optional: Identifier[Optional[int]] = 0
    pk_override_optional: Identifier[Optional[int]] = 0
    pk_override_none: Identifier[Optional[int]] = 0

    pk_fk_value: Identifier[ForeignModel]
    pk_fk_default_none: Identifier[Optional[ForeignModel]]
    pk_fk_default: Identifier[ForeignModel] = 0
    pk_fk_override: Identifier[ForeignModel] = 0
    pk_fk_default_optional: Identifier[Optional[ForeignModel]] = 0
    pk_fk_override_optional: Identifier[Optional[ForeignModel]] = 0
    pk_fk_override_none: Identifier[Optional[ForeignModel]] = 0

    fk_value: ForeignModel
    fk_default_none: Optional[ForeignModel]
    fk_default: ForeignModel = 0
    fk_override: ForeignModel = 0
    fk_default_optional: Optional[ForeignModel] = 0
    fk_override_optional: Optional[ForeignModel] = 0
    fk_override_none: Optional[ForeignModel] = 0

    field_value: int
    field_default_none: Optional[int]
    field_default: int = 0
    field_override: int = 0
    field_default_optional: Optional[int] = 0
    field_override_optional: Optional[int] = 0
    field_override_none: Optional[int] = 0


property_model = PropertyModel(
    pk_value=1,
    pk_override=1,
    pk_override_optional=1,
    pk_override_none=None,

    pk_fk_value=1,
    pk_fk_override=1,
    pk_fk_override_optional=1,
    pk_fk_override_none=None,

    fk_value=1,
    fk_override=1,
    fk_override_optional=1,
    fk_override_none=None,

    field_value=1,
    field_override=1,
    field_override_optional=1,
    field_override_none=None,
)

all_properties = [
    'pk_value',
    'pk_default_none',
    'pk_default',
    'pk_override',
    'pk_default_optional',
    'pk_override_optional',
    'pk_override_none',

    'pk_fk_value',
    'pk_fk_default_none',
    'pk_fk_default',
    'pk_fk_override',
    'pk_fk_default_optional',
    'pk_fk_override_optional',
    'pk_fk_override_none',

    'fk_value',
    'fk_default_none',
    'fk_default',
    'fk_override',
    'fk_default_optional',
    'fk_override_optional',
    'fk_override_none',

    'field_value',
    'field_default_none',
    'field_default',
    'field_override',
    'field_default_optional',
    'field_override_optional',
    'field_override_none'
]

@pytest.mark.parametrize('property_filter, expected', [
    (ALL, all_properties),
    (PK, {name for name in all_properties if 'pk_' in name}),
    (FK, {name for name in all_properties if 'fk_' in name}),
    (DEFAULT, {name for name in all_properties if 'default' in name}),
    (NONE, {name for name in all_properties if 'none' in name})
])
def test_basic_filter(property_filter: BasicPropertyFilter, expected: list[str]):
    assert property_filter.evaluate(property_model) == set(expected)

@pytest.mark.parametrize('property_filter', [ALL, PK, FK, DEFAULT, NONE])
def test_not_filter(property_filter: BasicPropertyFilter):
    normal_filter = property_filter.evaluate(property_model)
    not_filter = (~property_filter).evaluate(property_model)
    assert normal_filter.isdisjoint(not_filter)
    assert normal_filter | not_filter == set(all_properties)

@pytest.mark.parametrize('property_filter', [ALL, PK, FK, DEFAULT, NONE])
def test_or_filter(property_filter: BasicPropertyFilter):
    assert (property_filter | ~property_filter).evaluate(property_model) == set(all_properties)

@pytest.mark.parametrize('property_filter, expected', [
    (PK & FK, {name for name in all_properties if 'pk_fk_' in name}),
    (DEFAULT & NONE, {'pk_default_none', 'pk_fk_default_none', 'fk_default_none', 'field_default_none'}),
    (PK & DEFAULT, {'pk_default_none', 'pk_default', 'pk_default_optional', 'pk_fk_default_none', 'pk_fk_default', 'pk_fk_default_optional'}),
    (FK & NONE, {'pk_fk_default_none', 'pk_fk_override_none', 'fk_default_none', 'fk_override_none'}),
])
def test_and_filter(property_filter: AndFilter, expected: list[str]):
    assert property_filter.evaluate(property_model) == set(expected)

def test_evaluate_to_no_results():
    assert (~ALL).evaluate(property_model) == set()

@labeled_tests({
    'modified to None':
        (~DEFAULT & NONE, {'pk_override_none', 'pk_fk_override_none', 'fk_override_none', 'field_override_none'}),
    'existing relationships':
        (FK & ~NONE, {
            'pk_fk_value',
            'pk_fk_default',
            'pk_fk_override',
            'pk_fk_default_optional',
            'pk_fk_override_optional',

            'fk_value',
            'fk_default',
            'fk_override',
            'fk_default_optional',
            'fk_override_optional'
        }),
    'standard fields':
        (~PK & ~FK, {name for name in all_properties if 'field_' in name}),
    'keys':
        (PK | FK, {
            'pk_value',
            'pk_default_none',
            'pk_default',
            'pk_override',
            'pk_default_optional',
            'pk_override_optional',
            'pk_override_none',

            'pk_fk_value',
            'pk_fk_default_none',
            'pk_fk_default',
            'pk_fk_override',
            'pk_fk_default_optional',
            'pk_fk_override_optional',
            'pk_fk_override_none',

            'fk_value',
            'fk_default_none',
            'fk_default',
            'fk_override',
            'fk_default_optional',
            'fk_override_optional',
            'fk_override_none'
        }),
    'fields that are primary and non-null or foreign and default':
        (PK & ~NONE | FK & DEFAULT, {
            'pk_value',
            'pk_default',
            'pk_override',
            'pk_default_optional',
            'pk_override_optional',

            'pk_fk_value',
            'pk_fk_default_none',
            'pk_fk_default',
            'pk_fk_override',
            'pk_fk_default_optional',
            'pk_fk_override_optional',

            'fk_default_none',
            'fk_default',
            'fk_default_optional'
        }),
    'non-keys with a modified value that is not None':
        (~FK & ~PK & ~DEFAULT & ~NONE, {'field_value', 'field_override', 'field_override_optional'}),
    'default fields except unset optional unless it is pk fk or not a key at all':
        (DEFAULT & ~(NONE & DEFAULT & ~(PK & FK | ~(PK | FK))), {
            'pk_default',
            'pk_default_optional',

            'pk_fk_default_none',
            'pk_fk_default',
            'pk_fk_default_optional',

            'fk_default',
            'fk_default_optional',

            'field_default_none',
            'field_default',
            'field_default_optional',
        })
})
def test_mixed_filters(property_filter: PropertyFilter, expected: set[str]):
    assert property_filter.evaluate(property_model) == expected

@labeled_tests({
    'primary key properties':
        (PK, [
            'pk_value',
            'pk_default_none',
            'pk_default',
            'pk_override',
            'pk_default_optional',
            'pk_override_optional',
            'pk_override_none',

            'pk_fk_value',
            'pk_fk_default_none',
            'pk_fk_default',
            'pk_fk_override',
            'pk_fk_default_optional',
            'pk_fk_override_optional',
            'pk_fk_override_none'
        ]),
    'foreign but not primary key properties':
        (FK & ~PK, [
            'fk_value',
            'fk_default_none',
            'fk_default',
            'fk_override',
            'fk_default_optional',
            'fk_override_optional',
            'fk_override_none'
        ])
})
def test_get_property_names(property_filter: PropertyFilter, expected: list[str]):
    assert property_model.get_property_names(property_filter) == expected

def test_get_property_names__no_args():
    assert property_model.get_property_names() == all_properties

def test_get_property_names__multiple_args():
    assert property_model.get_property_names(~PK, ~FK, ~NONE | DEFAULT) == [
        'field_value',
        'field_default_none',
        'field_default',
        'field_override',
        'field_default_optional',
        'field_override_optional'
    ]

@labeled_tests({
    'standard properties':
        (~(PK | FK), {
            'field_value': 1,
            'field_default_none': None,
            'field_default': 0,
            'field_override': 1,
            'field_default_optional': 0,
            'field_override_optional': 1,
            'field_override_none': None
        }),
    'no properties':
        (PK & ~PK, dict())
})
def test_get_property_values(property_filter: PropertyFilter, expected: dict[str, Any]):
    assert property_model.get_property_values(property_filter) == expected



from datetime import date

import pytest

from daomodel.base_service import SingleModelService, SOURCE_VALUE, DESTINATION_VALUE, BaseService
from daomodel.list_util import longest
from daomodel.testing import TestDAOFactory
from tests.school_models import Staff, Student, Book


def setup_staff(daos: TestDAOFactory) -> tuple[Staff, Staff]:
    dao = daos[Staff]
    ed = dao.create_with(id=1, name='Ed', hire_date=date(2023, 6, 15))
    edward = dao.create_with(id=2, name='Edward', hire_date=date(2024, 8, 20))
    return ed, edward


def test_merge(daos: TestDAOFactory):
    ed, edward = setup_staff(daos)
    SingleModelService(daos, Staff).merge(ed, 2, name=longest, hire_date=min)
    daos.assert_in_db(Staff, 2, name='Edward', hire_date=date(2023, 6, 15))
    daos.assert_not_in_db(Staff, 1)


def test_merge__source_destination_values(daos: TestDAOFactory):
    ed, edward = setup_staff(daos)
    service = SingleModelService(daos, Staff)
    service.merge(edward, 1, name=DESTINATION_VALUE, hire_date=SOURCE_VALUE)
    daos.assert_in_db(Staff, 1, name='Ed', hire_date=date(2024, 8, 20))
    daos.assert_not_in_db(Staff, 2)


def test_merge__mismatched_model_type(daos: TestDAOFactory):
    setup_staff(daos)
    service = SingleModelService(daos, Staff)
    student = daos[Student].create_with(id=100, name='Student', gender='m')

    with pytest.raises(TypeError):
        service.merge(student, 1)


def test_merge__redirect(daos: TestDAOFactory):
    dao = daos[Student]
    tim = dao.create_with(id=1, name='Tim')
    timothy = dao.create_with(id=2, name='Timothy')
    daos[Book].create_with(name='Biology 101', subject='Science', owner=tim.id)

    SingleModelService(daos, Student).merge(tim, 2, name=longest)
    daos.assert_in_db(Book, 'Biology 101', subject='Science', owner=timothy.id)


def test_dao(daos: TestDAOFactory):
    service = SingleModelService(daos, Staff)
    staff = service.dao.create_with(id=3, name='Alice', hire_date=date(2023, 1, 15))

    daos.assert_in_db(Staff, 3, name='Alice', hire_date=date(2023, 1, 15))

    staff.name = 'Alicia'
    service.dao.commit(staff)

    daos.assert_in_db(Staff, 3, name='Alicia', hire_date=date(2023, 1, 15))


def test_bulk_update(daos: TestDAOFactory):
    service = SingleModelService(daos, Student)
    fee = service.dao.create_with(id=1, name='Fee')
    fi = service.dao.create_with(id=2, name='Fi')
    fo = service.dao.create_with(id=3, name='Fo')
    fum = service.dao.create_with(id=4, name='Fum')

    daos.assert_in_db(Student, 1, name='Fee', active=True)
    daos.assert_in_db(Student, 2, name='Fi', active=True)
    daos.assert_in_db(Student, 3, name='Fo', active=True)
    daos.assert_in_db(Student, 4, name='Fum', active=True)

    service.bulk_update([fee, fi, fo, fum], active=False)

    daos.assert_in_db(Student, 1, name='Fee', active=False)
    daos.assert_in_db(Student, 2, name='Fi', active=False)
    daos.assert_in_db(Student, 3, name='Fo', active=False)
    daos.assert_in_db(Student, 4, name='Fum', active=False)


def test_bulk_update__multiple_fields(daos: TestDAOFactory):
    dao = daos[Student]
    tim = dao.create_with(id=1, name='Tim')
    tom = dao.create_with(id=2, name='Tom')
    tam = dao.create_with(id=3, name='Tam', gender='f')

    BaseService(daos).bulk_update([tim, tom], gender='m', active=False)

    daos.assert_in_db(Student, 1, name='Tim', gender='m', active=False)
    daos.assert_in_db(Student, 2, name='Tom', gender='m', active=False)
    daos.assert_in_db(Student, 3, name='Tam', gender='f', active=True)


def test_bulk_update__multiple_models(daos: TestDAOFactory):
    staff = daos[Staff].create_with(id=1, name='Mr. Staff', hire_date=date(2020, 2, 20))
    student = daos[Student].create_with(id=2, name='Ms. Student')

    BaseService(daos).bulk_update([staff, student], name='Pending Removal')

    daos.assert_in_db(Staff, 1, name='Pending Removal')
    daos.assert_in_db(Student, 2, name='Pending Removal')


def test_bulk_update__empty(daos: TestDAOFactory):
    service = SingleModelService(daos, Staff)
    service.bulk_update([], name='Should not be applied')


def test_bulk_update__non_applicable_fields(daos: TestDAOFactory):
    staff1 = daos[Staff].create_with(id=1, name='Mr. Staff', hire_date=date(2020, 2, 20))
    student1 = daos[Student].create_with(id=2, name='Ms. Student')
    staff2 = daos[Staff].create_with(id=3, name='Sir Staff', hire_date=date(2024, 2, 20))
    student2 = daos[Student].create_with(id=4, name='Mr. Student')

    BaseService(daos).bulk_update(
        [staff1, staff2, student1, student2],
        active=False,
        hire_date=date(2024, 2, 20)
    )

    daos.assert_in_db(Staff, 1, hire_date=date(2024, 2, 20))
    daos.assert_in_db(Staff, 3, hire_date=date(2024, 2, 20))
    daos.assert_in_db(Student, 2, active=False)
    daos.assert_in_db(Student, 4, active=False)


def test_bulk_update__rollback(daos: TestDAOFactory):
    staff1 = daos[Staff].create_with(id=1, name='Mr. Staff', hire_date=date(2020, 2, 20))
    staff2 = daos[Staff].create_with(id=3, name='Sir Staff', hire_date=date(2024, 2, 20))

    with pytest.raises(Exception):
        BaseService(daos).bulk_update(
            [staff1, staff2],
            hire_date='Feb. 20th'
        )

    daos.assert_in_db(Staff, 1, hire_date=date(2020, 2, 20))
    daos.assert_in_db(Staff, 3, hire_date=date(2024, 2, 20))




from typing import TypeVar, Any, Iterable, Optional


T = TypeVar('T', bound=Any)


def ensure_iter(elements: Iterable[T] | T) -> Iterable[T] | list[T]:
    """Ensures that the provided argument is iterable.

    Single, non-Iterable items are converted to a single-item list.
    In this context, a str is not considered to be Iterable.

    :param elements: The input that may or may not be Iterable
    :return: The provided Iterable or a single item list
    """
    if not isinstance(elements, Iterable) or type(elements) is str:
        elements = [elements]
    return elements


def in_order(original: Iterable[T], order: list[T]) -> list[T]:
    """Returns provided items as an ordered list.

    Repeated items will be deduplicated.
    Items not defined within the order will be excluded.
    The order is allowed to contain extraneous items that aren't applicable to the provided items.

    :param original: The (likely unordered) collection of items
    :param order: The defined order of items
    :return: a new list of the items following the defined order
    """
    return [item for item in order if item in original]


def strip_whitespace(values: list[T]) -> list[T]:
    """Trim leading and trailing whitespace from strings in a list."""
    return [value.strip() for value in values]


def exclude_falsy(values: list[T]) -> list[T]:
    """Returns a list of only the truthy values (excluding None, '', 0, False, etc)."""
    return [v for v in values if v]


def dedupe(original: list[T], keep_last=False) -> list[T]:
    """Creates a filtered copy of a list that does not include duplicates.

    :param original: The list to filter
    :param keep_last: True to keep the last occurrence of a duplicate, otherwise the first occurrence will be kept
    :return: a new list that maintains order but is guaranteed to have no duplicates
    """
    if keep_last:
        return dedupe(original[::-1])[::-1]
    else:
        return list(dict.fromkeys(original))


def most_frequent(values: list[T]) -> T:
    """Determines the most frequently occurring value within the provided iterable.

    In the case of a tie (multiple values with the same highest frequency),
    the function returns the first value encountered with that frequency.

    :param values: A list containing values to evaluate
    :return: The most common value from the provided iterable
    """
    return max(values, key=values.count)


def longest(values: list[str]) -> Optional[str]:
    """Finds the longest string in a list of strings.

    In the case of a tie (multiple values of the same length),
    the function returns the first value encountered with that length.

    :param values: A list containing strings to check
    :return: The longest string from the provided list.
    """
    try:
        return max(values, key=len)
    except ValueError:
        return None


def first_str_with(substring: str, strings: list[str]) -> Optional[str]:
    """Returns the first str that contains the given substring.

    :param substring: The substring to search for.
    :param strings: One or more strings to check.
    :return: The first string (if any) that contains the substring.
    """
    for s in strings:
        if substring in s:
            return s
    return None


def first(values: list[T]) -> Optional[T]:
    """Returns the first truthy value.

    :param values: Values to evaluate
    :return: First truthy value found, or None if no truthy values exist
    """
    for value in values:
        if value:
            return value
    return None


def last(values: list[T]) -> Optional[T]:
    """Returns the last truthy value.

    :param values: Values to evaluate
    :return: Last truthy value found, or None if no truthy values exist
    """
    return first(values[::-1])


from sqlmodel import desc

from daomodel import UnsearchableError
from daomodel.dao import SearchResults
from daomodel.search_util import LessThan, GreaterThan, GreaterThanEqualTo, LessThanEqualTo, Between, \
    AnyOf, NoneOf, IsSet, is_set, NotSet, not_set
from daomodel.util import MissingInput
from tests.school_models import *


first_page = SearchResults(page_one, total=len(all_students), page=1, per_page=5)
subsequent_page = SearchResults(page_two, total=len(all_students), page=2, per_page=5)
last_page = SearchResults(page_three, total=len(all_students), page=3, per_page=5)
large_page = SearchResults(all_students, total=len(all_students), page=1, per_page=15)
no_page = SearchResults(all_students)


@pytest.mark.parametrize('results, expected_start, expected_end, expected_total_pages', [
    (first_page, 1, 5, 3),
    (large_page, 1, 13, 1),
    (subsequent_page, 6, 10, 3),
    (last_page, 11, 13, 3),
    (no_page, 1, 13, 1)
])
def test_page_start_end_total(results: SearchResults, expected_start: int, expected_end: int, expected_total_pages: int):
    assert results.page_start == expected_start
    assert results.page_end == expected_end
    assert results.total_pages == expected_total_pages


def test_find__all(student_dao: DAO):
    assert student_dao.find() == SearchResults(all_students)


def test_first__multiple_results(student_dao: DAO):
    assert student_dao.find().first() == all_students[0]


def test_only__multiple_results(student_dao: DAO):
    with pytest.raises(ValueError):
        student_dao.find().only()


def test_find__single_result(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create(100)
    assert dao.find() == SearchResults([Student(id=100)])


def test_first__single_result(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create(100)
    assert dao.find().first() == Student(id=100)


def test_only__single_result(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create(100)
    assert dao.find().only() == Student(id=100)


def test_find__no_results(daos: TestDAOFactory):
    assert daos[Student].find() == SearchResults([])


def test_first__no_results(daos: TestDAOFactory):
    assert daos[Student].find().first() is None


def test_only__no_results(daos: TestDAOFactory):
    with pytest.raises(ValueError):
        daos[Student].find().only()


def test_find__limit(student_dao: DAO):
    assert student_dao.find(_per_page=5) == first_page


def test_find__fewer_results_than_limit(student_dao: DAO):
    assert student_dao.find(_per_page=15) == large_page


def test_find__subsequent_page(student_dao: DAO):
    assert student_dao.find(_per_page=5, _page=2) == subsequent_page


def test_find__last_page(student_dao: DAO):
    assert student_dao.find(_per_page=5, _page=3) == last_page


def test_find__undefined_page_size(daos: TestDAOFactory):
    with pytest.raises(MissingInput):
        daos[Student].find(_page=1)


def test_find__filter_by_single_property(student_dao: DAO):
    assert student_dao.find(id=106) == SearchResults([Student(id=106)])


def test_find__filter_by_bool_property(student_dao: DAO):
    assert student_dao.find(active=True) == SearchResults(active)
    assert student_dao.find(active=False) == SearchResults(inactive)


def test_find__unsearchable_property(daos: TestDAOFactory):
    with pytest.raises(UnsearchableError):
        daos[Person].find(ssn=32)


def test_find__invalid_property(daos: TestDAOFactory):
    with pytest.raises(UnsearchableError):
        daos[Student].find(sex='m')


def test_find__filter_by_multiple_properties(student_dao: DAO):
    assert student_dao.find(gender='f', active=True) == SearchResults(active_females)


def test_find__is_set(student_dao: DAO):
    #assert student_dao.find(is_set_=Student.gender) == SearchResults(having_gender)
    assert student_dao.find(gender=IsSet()) == SearchResults(having_gender)
    assert student_dao.find(gender=is_set) == SearchResults(having_gender)


@pytest.mark.skip(reason='Not yet implemented')
def test_find__is_set_foreign_property(student_dao: DAO):
    assert student_dao.find(is_set_=Book.name) == SearchResults(having_book)


def test_find__is_set_unsearchable(student_dao: DAO):
    with pytest.raises(UnsearchableError):
        student_dao.find(is_set_=Hall.floor)


def test_find__not_set(student_dao: DAO):
    #assert student_dao.find(not_set_=Student.name) == SearchResults(not_having_name)
    assert student_dao.find(name=NotSet()) == SearchResults(not_having_name)
    assert student_dao.find(name=not_set) == SearchResults(not_having_name)


@pytest.mark.skip(reason='Not yet implemented')
def test_find__not_set_foreign_property(student_dao: DAO):
    assert student_dao.find(not_set_=Locker.number) == SearchResults(not_having_locker)


def test_find__condition_operator_unsearchable(person_dao: DAO):
    with pytest.raises(UnsearchableError):
        person_dao.find(ssn=is_set)


def test_find__filter_by_0_value(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create(0)
    assert dao.find(id=0) == SearchResults([Student(id=0)])


def test_find__filter_by_foreign_property(school_dao: DAO):
    assert school_dao.find(**{'book.subject': 'Math'}) == SearchResults(having_math_book)


def test_find__filter_by_multiple_foreign_property(school_dao: DAO):
    filters = {'book.name': 'Calculus', 'book.subject': 'Math'}
    assert school_dao.find(**filters) == SearchResults([Student(id=103)])


def test_find__filter_by_different_foreign_tables(school_dao: DAO):
    filters = {'book.name': 'Biology 101', 'locker.number': 1101}
    assert school_dao.find(**filters) == SearchResults([Student(id=100)])


def test_find__filter_by_nested_foreign_property(school_dao: DAO):
    assert school_dao.find(**{'hall.color': 'blue'}) == SearchResults(in_blue_hall)


def test_find__filter_by_gt(school_dao: DAO):
    expected = [Student(id=109), Student(id=110), Student(id=111), Student(id=112)]
    assert school_dao.find(id=GreaterThan(108)) == SearchResults(expected)


def test_find__filter_by_gteq(school_dao: DAO):
    expected = [Student(id=108), Student(id=109), Student(id=110), Student(id=111), Student(id=112)]
    assert school_dao.find(id=GreaterThanEqualTo(108)) == SearchResults(expected)


def test_find__filter_by_lt(school_dao: DAO):
    expected = [Student(id=100), Student(id=101), Student(id=102), Student(id=103)]
    assert school_dao.find(id=LessThan(104)) == SearchResults(expected)


def test_find__filter_by_lteq(school_dao: DAO):
    expected = [Student(id=100), Student(id=101), Student(id=102), Student(id=103), Student(id=104)]
    assert school_dao.find(id=LessThanEqualTo(104)) == SearchResults(expected)


def test_find__filter_by_between(school_dao: DAO):
    expected = [Student(id=104), Student(id=105), Student(id=106), Student(id=107), Student(id=108)]
    assert school_dao.find(id=Between(104, 108)) == SearchResults(expected)


def test_find__filter_by_any_of(person_dao: DAO):
    expected = [
        Person(name='Greg', age=31),
        Person(name='John', age=23),
        Person(name='John', age=45)
    ]
    assert person_dao.find(name=AnyOf('John', 'Greg')) == SearchResults(expected)


def test_find__filter_by_none_of(person_dao: DAO):
    expected = [
        Person(name='Mike', age=18),
        Person(name='Mike', age=25),
        Person(name='Paul', age=25)
    ]
    assert person_dao.find(name=NoneOf('John', 'Joe', 'Greg')) == SearchResults(expected)


def test_find__default_order(person_dao: DAO):
    assert person_dao.find() == SearchResults(pk_ordered)


def test_find__specified_order(person_dao: DAO):
    assert person_dao.find(_order=Person.age) == SearchResults(age_ordered)


def test_find__reverse_order(student_dao: DAO):
    assert student_dao.find(_order=desc(Student.id)) == SearchResults(list(reversed(all_students)))


def test_find__order_without_table(person_dao: DAO):
    assert person_dao.find(_order='age') == SearchResults(age_ordered)


def test_find__order_by_multiple_properties(student_dao: DAO):
    ordered = [
        Student(id=111),
        Student(id=106),
        Student(id=107),
        Student(id=102),
        Student(id=110),
        Student(id=112),
        Student(id=109),
        Student(id=108),
        Student(id=105),
        Student(id=101),
        Student(id=103),
        Student(id=104),
        Student(id=100)
    ]
    order = (Student.active, Student.gender, desc(Student.name))
    assert student_dao.find(_order=order) == SearchResults(ordered)


def test_find__order_by_foreign_property(school_dao: DAO):
    ordered = [Student(id=101), Student(id=100), Student(id=103), Student(id=102)]
    assert school_dao.find(_order=Book.name) == SearchResults(ordered)


def test_find__order_by_nested_foreign_property(school_dao: DAO):
    ordered = [
        Student(id=107),
        Student(id=110),
        Student(id=102),
        Student(id=100),
        Student(id=108),
        Student(id=103),
        Student(id=104),
        Student(id=109),
        Student(id=101),
        Student(id=106),
        Student(id=111),
        Student(id=105)
    ]
    assert school_dao.find(_order=Hall.color) == SearchResults(ordered)


def test_find__order_by_unsearchable(daos: TestDAOFactory):
    with pytest.raises(UnsearchableError):
        daos[Person].find(_order=Person.ssn)


def test_find__duplicate(person_dao: DAO):
    assert person_dao.find(_duplicate=Person.name) == SearchResults(duplicated_names)


def test_find__duplicate_foreign_property(school_dao: DAO):
    assert school_dao.find(_duplicate=Book.subject) == SearchResults([Student(id=102), Student(id=103)])


def test_find__duplicate_unsearchable(daos: TestDAOFactory):
    with pytest.raises(UnsearchableError):
        daos[Person].find(_duplicate=Person.ssn)


def test_find__unique(person_dao: DAO):
    assert person_dao.find(_unique=Person.name) == SearchResults(unique_names)


def test_find__unique_foreign_property(school_dao: DAO):
    assert school_dao.find(_unique=Book.subject) == SearchResults([Student(id=100), Student(id=101)])


def test_find__unique_unsearchable(daos: TestDAOFactory):
    with pytest.raises(UnsearchableError):
        daos[Person].find(_unique=Person.ssn)


def test_find__duplicate_and_unique(person_dao: DAO):
    expected = [
        Person(name='John', age=23),
        Person(name='John', age=45),
        Person(name='Mike', age=18)
    ]
    assert person_dao.find(_duplicate=Person.name, _unique=Person.age) == SearchResults(expected)


def test_search_results__iter(student_dao: DAO):
    student_id = 100
    for student in student_dao.find():
        assert student.id == student_id
        student_id += 1


def test_search_results__eq_hash(student_dao: DAO):
    first = student_dao.find()
    second = student_dao.find()
    assert first == second
    assert hash(first) == hash(second)


def test_search_results__eq_hash__different_order(student_dao: DAO):
    first = student_dao.find(_order=Student.name)
    second = student_dao.find(_order=desc(Student.name))
    assert first != second
    assert hash(first) != hash(second)


def test_search_results__str(student_dao: DAO):
    results = student_dao.find()
    assert str(results) == str(list(results))


def test_search_results__str__page(student_dao: DAO):
    result = str(student_dao.find(_page=2, _per_page=5))
    assert result.split('[')[0] == 'Page 2 of 3; 6-10 of 13 results '



from datetime import date
from typing import Optional, Any

import pytest

from daomodel import DAOModel
from daomodel.change_set import Preference, ChangeSet, MergeSet, Resolved, Unresolved
from daomodel.dao import Conflict
from daomodel.fields import Identifier
from daomodel.list_util import most_frequent
from daomodel.testing import labeled_tests


class CalendarEvent(DAOModel, table=True):
    title: Identifier[str]
    day: date
    time: Optional[str] = 'All Day'
    location: Optional[str]
    description: Optional[str]


dads_entry = CalendarEvent(
    title='Family Picnic',
    day=date(2025, 6, 20),
    time='11:00 AM',
    location='Central Park',
    description='Annual family picnic with games and BBQ.'
)
moms_entry = CalendarEvent(
    title='Family Picnic',
    day=date(2025, 6, 20),
    time='12:00 PM',
    location='Central Park',
    description='Picnic with family and friends, do not forget the salads!'
)
sons_entry = CalendarEvent(
    title='Family Picnic',
    day=date(2025, 6, 19),
    time='12:00 PM',
    description='Bring your football and frisbee!'
)
daughters_entry = CalendarEvent(
    title='Family Picnic',
    day=date(2025, 6, 20),
    time='All Day',
    location='The Park'
)
unrelated_entry = CalendarEvent(
    title='Dentist Appointment',
    day=date(2025, 7, 1)
)


rules = {
    'title': Preference.NOT_APPLICABLE,
    'day': max,
    'time': min,
    'location_conflict': Preference.LEFT,
    'description_conflict': '\n\n'.join
}
merge_rules = {**rules, 'location_conflict': most_frequent, 'description_conflict': Preference.LEFT}


def test_change_set():
    assert 'title' not in ChangeSet(dads_entry, unrelated_entry)


def test_change_set__include_pk():
    assert 'title' in ChangeSet(dads_entry, unrelated_entry, include_pk=True)


def test_get_baseline_get_target():
    change_set = ChangeSet(dads_entry, moms_entry)
    assert change_set.get_baseline('time') == change_set.get_left('time') == '11:00 AM'
    assert change_set.get_target('time') == change_set.get_right('time') == '12:00 PM'


@labeled_tests({
    'ChangeSet has target value':
        (ChangeSet(dads_entry, moms_entry), 'time', True),
    'ChangeSet does not have target value':
        (ChangeSet(dads_entry, sons_entry), 'location', False),
    'Single Target MergeSet has target value':
        (MergeSet(dads_entry, moms_entry), 'time', True),
    'Single Target MergeSet does not have target value':
        (MergeSet(dads_entry, sons_entry), 'location', False),
    'Multiple Target MergeSet has all target value':
        (MergeSet(dads_entry, moms_entry, sons_entry, daughters_entry), 'time', True),
    'Multiple Target MergeSet has some target value':
        (MergeSet(dads_entry, moms_entry, sons_entry, daughters_entry), 'description', True),
    'Multiple Target MergeSet has no target value':
        (MergeSet(dads_entry, sons_entry, unrelated_entry), 'location', False),
})
def test_has_target_value(change_set: ChangeSet, field: str, expected: bool):
    assert change_set.has_target_value(field) is expected


@labeled_tests({
    'ChangeSet':
        (ChangeSet(dads_entry, moms_entry), 'time', ['11:00 AM', '12:00 PM']),
    'ChangeSet None Baseline':
        (ChangeSet(daughters_entry, sons_entry), 'description', [None, 'Bring your football and frisbee!']),
    'Single Target MergeSet':
        (MergeSet(dads_entry, sons_entry), 'day', [date(2025, 6, 20), date(2025, 6, 19)]),
    'Multiple Target MergeSet':
        (MergeSet(dads_entry, moms_entry, daughters_entry), 'time', [
            '11:00 AM',
            '12:00 PM',
            'All Day'
        ]),
    'Multiple Target MergeSet with some unchanged values':
        (MergeSet(dads_entry, moms_entry, daughters_entry), 'location', [
            'Central Park',
            'Central Park',
            'The Park'
        ]),
    'Multiple Target MergeSet with None value':
        (MergeSet(dads_entry, moms_entry, sons_entry, daughters_entry), 'description', [
            'Annual family picnic with games and BBQ.',
            'Picnic with family and friends, do not forget the salads!',
            'Bring your football and frisbee!',
            None
        ]),
    'Multiple Target MergeSet with None Baseline':
        (MergeSet(sons_entry, dads_entry, daughters_entry), 'location', [
            None,
            'Central Park',
            'The Park'
        ]),
    'Multiple Target MergeSet with all target matching':
        (MergeSet(sons_entry, dads_entry, moms_entry, daughters_entry), 'day', [
            date(2025, 6, 19),
            date(2025, 6, 20),
            date(2025, 6, 20),
            date(2025, 6, 20)
        ]),
})
def test_all_values(change_set: ChangeSet, field: str, expected: list[Any]):
    assert change_set.all_values(field) == expected


def test_get_resolution():
    change_set = ChangeSet(dads_entry, moms_entry, **rules)
    change_set.resolve_preferences()
    assert (change_set.get_resolution('description') ==
            'Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!')
    change_set = ChangeSet(moms_entry, dads_entry, **rules)
    change_set.resolve_preferences()
    assert (change_set.get_resolution('description') ==
            'Picnic with family and friends, do not forget the salads!\n\nAnnual family picnic with games and BBQ.')


def test_get_resolution__unresolved():
    assert ChangeSet(dads_entry, moms_entry).get_resolution('time') == '12:00 PM'
    assert ChangeSet(moms_entry, dads_entry).get_resolution('time') == '11:00 AM'


get_preferred_tests = {
    'left': [
        (dads_entry, sons_entry, 'location', Preference.LEFT),
        (dads_entry, daughters_entry, 'time', Preference.LEFT),
        (dads_entry, daughters_entry, 'description', Preference.LEFT),
        (sons_entry, daughters_entry, 'time', Preference.LEFT),
        (sons_entry, daughters_entry, 'description', Preference.LEFT)
    ],
    'right': [
        (sons_entry, moms_entry, 'location', Preference.RIGHT),
        (daughters_entry, moms_entry, 'time', Preference.RIGHT),
        (daughters_entry, moms_entry, 'description', Preference.RIGHT),
        (sons_entry, daughters_entry, 'location', Preference.RIGHT)
    ],
    'both': [
        (dads_entry, moms_entry, 'time', Preference.BOTH),
        (dads_entry, moms_entry, 'description', Preference.BOTH),
        (moms_entry, sons_entry, 'day', Preference.BOTH)
    ]
}
@labeled_tests(get_preferred_tests)
def test_get_preferred(baseline: CalendarEvent, target: CalendarEvent, field: str, expected: Preference):
    assert ChangeSet(baseline, target).get_preferred(field) == expected


@labeled_tests({
    'resolve by preference': [
        (ChangeSet(dads_entry, daughters_entry, location_conflict=Preference.LEFT), 'location', Preference.LEFT),
        (ChangeSet(dads_entry, daughters_entry, location_conflict=Preference.RIGHT), 'location', Preference.RIGHT),
        (ChangeSet(dads_entry, moms_entry, description_conflict=Preference.NOT_APPLICABLE), 'description', Preference.NOT_APPLICABLE),
        (ChangeSet(dads_entry, moms_entry, description_conflict=Preference.NEITHER), 'description', Preference.NEITHER),
        (MergeSet(dads_entry, daughters_entry, location_conflict=Preference.LEFT), 'location', Preference.LEFT),
        (MergeSet(dads_entry, moms_entry, sons_entry, day_conflict=Preference.RIGHT), 'day', Preference.RIGHT),
        (MergeSet(dads_entry, moms_entry, description_conflict=Preference.NEITHER), 'description', Preference.NEITHER)
    ],
    'resolve by comparison': [
        (ChangeSet(dads_entry, sons_entry, day_conflict=max), 'day', Preference.LEFT),
        (ChangeSet(dads_entry, sons_entry, day_conflict=min), 'day', Preference.RIGHT)
    ],
    'static resolve': [
        (ChangeSet(dads_entry, moms_entry, time_conflict='11:00 AM'), 'time', Preference.LEFT),
        (ChangeSet(dads_entry, moms_entry, time_conflict='12:00 PM'), 'time', Preference.RIGHT),
        (ChangeSet(dads_entry, moms_entry, time_conflict='11:30 AM'), 'time', '11:30 AM'),
        (ChangeSet(dads_entry, moms_entry, time_conflict=None), 'time', None)
    ],
    'none value preferred':
        (MergeSet(dads_entry, sons_entry, daughters_entry, description_conflict=None), 'description', (Preference.RIGHT, 1)),
    'both': [
        (ChangeSet(dads_entry, moms_entry, description_conflict='\n\n'.join), 'description',
         'Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!'),
        (MergeSet(dads_entry, moms_entry, description_conflict='\n\n'.join), 'description',
         'Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!'),
        (MergeSet(dads_entry, moms_entry, sons_entry, description_conflict='\n\n'.join), 'description',
         'Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!\n\nBring your football and frisbee!')
    ],
    'default conflict resolution': [
        (ChangeSet(dads_entry, moms_entry, default_conflict=Preference.LEFT), 'time', Preference.LEFT),
        (MergeSet(dads_entry, moms_entry, default_conflict=Preference.LEFT), 'time', Preference.LEFT),
        (MergeSet(dads_entry, moms_entry, sons_entry, default_conflict=Preference.LEFT), 'time', Preference.LEFT)
    ],
    'MergeSet resolve by comparison': [
        (MergeSet(dads_entry, sons_entry, day_conflict=max), 'day', Preference.LEFT),
        (MergeSet(dads_entry, sons_entry, day_conflict=min), 'day', Preference.RIGHT),
        (MergeSet(dads_entry, moms_entry, sons_entry, daughters_entry, day_conflict=max), 'day', Preference.LEFT),
        (MergeSet(dads_entry, moms_entry, sons_entry, daughters_entry, day_conflict=most_frequent), 'day', Preference.LEFT),
        (MergeSet(dads_entry, moms_entry, sons_entry, daughters_entry, time_conflict=most_frequent), 'time', (Preference.RIGHT, 0))
    ],
    'MergeSet static resolve': [
        (MergeSet(dads_entry, moms_entry, time_conflict='11:00 AM'), 'time', Preference.LEFT),
        (MergeSet(dads_entry, sons_entry, daughters_entry, time_conflict='11:00 AM'), 'time', Preference.LEFT),
        (MergeSet(dads_entry, moms_entry, time_conflict='12:00 PM'), 'time', Preference.RIGHT),
        (MergeSet(dads_entry, moms_entry, daughters_entry, time_conflict='All Day'), 'time', (Preference.RIGHT, 1)),
        (MergeSet(dads_entry, moms_entry, sons_entry, time_conflict='11:30 AM'), 'time', '11:30 AM')
    ],
    'multiple target matches':
        (MergeSet(dads_entry, moms_entry, sons_entry, time_conflict='12:00 PM'), 'time', (Preference.RIGHT, 0)),
    'matches left and right':
        (MergeSet(sons_entry, dads_entry, moms_entry, daughters_entry, time_conflict='12:00 PM'), 'time', Preference.LEFT)
})
def test_resolve_conflict(change_set: ChangeSet, field: str, expected: Preference|tuple[Preference.RIGHT, int]|Any):
    assert change_set.resolve_conflict(field) == expected


def test_resolve_conflict__missing():
    with pytest.raises(Conflict):
        ChangeSet(dads_entry, moms_entry).resolve_conflict('time')


@labeled_tests({
    'dad => mom':
        (dads_entry, moms_entry, {
            'description': (
                    'Annual family picnic with games and BBQ.',
                    Resolved('Picnic with family and friends, do not forget the salads!',
                             'Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!')
            )
        }),
    'dad => son':
        (dads_entry, sons_entry, {
            'description': (
                    'Annual family picnic with games and BBQ.',
                    Resolved('Bring your football and frisbee!',
                             'Annual family picnic with games and BBQ.\n\nBring your football and frisbee!')
            )
        }),
    'dad => daughter':
        (dads_entry, daughters_entry, {}),
    'mom => dad':
        (moms_entry, dads_entry, {
            'time': ('12:00 PM', '11:00 AM'),
            'description': (
                    'Picnic with family and friends, do not forget the salads!',
                    Resolved('Annual family picnic with games and BBQ.',
                             'Picnic with family and friends, do not forget the salads!\n\nAnnual family picnic with games and BBQ.')
            )
        }),
    'mom => son':
        (moms_entry, sons_entry, {
            'description': (
                    'Picnic with family and friends, do not forget the salads!',
                    Resolved('Bring your football and frisbee!',
                             'Picnic with family and friends, do not forget the salads!\n\nBring your football and frisbee!')
            )
        }),
    'mom => daughter':
        (moms_entry, daughters_entry, {}),
    'son => dad':
        (sons_entry, dads_entry, {
            'day': (date(2025, 6, 19), date(2025, 6, 20)),
            'time': ('12:00 PM', '11:00 AM'),
            'location': (None, 'Central Park'),
            'description': (
                    'Bring your football and frisbee!',
                    Resolved('Annual family picnic with games and BBQ.',
                             'Bring your football and frisbee!\n\nAnnual family picnic with games and BBQ.')
            )
        }),
    'son => mom':
        (sons_entry, moms_entry, {
            'day': (date(2025, 6, 19), date(2025, 6, 20)),
            'location': (None, 'Central Park'),
            'description': (
                    'Bring your football and frisbee!',
                    Resolved('Picnic with family and friends, do not forget the salads!',
                             'Bring your football and frisbee!\n\nPicnic with family and friends, do not forget the salads!')
            )
        }),
    'son => daughter':
        (sons_entry, daughters_entry, {
            'day': (date(2025, 6, 19), date(2025, 6, 20)),
            'location': (None, 'The Park')
        }),
    'daughter => dad':
        (daughters_entry, dads_entry, {
            'time': ('All Day', '11:00 AM'),
            'description': (None, 'Annual family picnic with games and BBQ.')
        }),
    'daughter => mom':
        (daughters_entry, moms_entry, {
            'time': ('All Day', '12:00 PM'),
            'description': (None, 'Picnic with family and friends, do not forget the salads!')
        }),
    'daughter => son':
        (daughters_entry, sons_entry, {
            'time': ('All Day', '12:00 PM'),
            'description': (None, 'Bring your football and frisbee!')
        })
})
def test_resolve_preferences(baseline: CalendarEvent, target: CalendarEvent, expected: dict[str, tuple[Any, Any|Resolved]]):
    change_set = ChangeSet(baseline, target, **rules)
    change_set.resolve_preferences()
    assert change_set == expected


def test_resolve_preferences__unresolved():
    change_set = ChangeSet(dads_entry, unrelated_entry, include_pk=True, **rules)
    change_set.resolve_preferences()
    assert change_set['title'] == ('Family Picnic', Unresolved('Dentist Appointment'))


def test_resolve_preferences__chained():
    assert ChangeSet(moms_entry, dads_entry, **rules).resolve_preferences().get_baseline('time') == '12:00 PM'


def test_resolve_preferences__conflict():
    with pytest.raises(Conflict):
        ChangeSet(dads_entry, moms_entry).resolve_preferences()


@labeled_tests({
    'dad => mom':
        (dads_entry, moms_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!'
        )),
    'dad => son':
        (dads_entry, sons_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Bring your football and frisbee!\n\nAnnual family picnic with games and BBQ.'
        )),
    'dad => daughter':
        (dads_entry, daughters_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.'
        )),
    'mom => dad':
        (moms_entry, dads_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.\n\nPicnic with family and friends, do not forget the salads!'
        )),
    'mom => son':
        (moms_entry, sons_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Bring your football and frisbee!\n\nPicnic with family and friends, do not forget the salads!'
        )),
    'mom => daughter':
        (moms_entry, daughters_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Picnic with family and friends, do not forget the salads!'
        )),
    'son => dad':
        (sons_entry, dads_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.\n\nBring your football and frisbee!'
        )),
    'son => mom':
        (sons_entry, moms_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Picnic with family and friends, do not forget the salads!\n\nBring your football and frisbee!'
        )),
    'son => daughter':
        (sons_entry, daughters_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Bring your football and frisbee!'
        )),
    'daughter => dad':
        (daughters_entry, dads_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.'
        )),
    'daughter => mom':
        (daughters_entry, moms_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Picnic with family and friends, do not forget the salads!'
        )),
    'daughter => son':
        (daughters_entry, sons_entry, CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='12:00 PM',
            location='Central Park',
            description='Bring your football and frisbee!'
        )),
})
def test_apply(baseline: CalendarEvent, target: CalendarEvent, expected: CalendarEvent):
    change_set = ChangeSet(baseline.model_copy(), target, **rules)
    change_set.resolve_preferences()
    assert change_set.apply() == expected


@labeled_tests({
    'dad <= mom, son, daughter':
        (dads_entry, [moms_entry, sons_entry, daughters_entry], CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.'
        )),
    'mom <= son, daughter, dad':
        (moms_entry, [sons_entry, daughters_entry, dads_entry], CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Picnic with family and friends, do not forget the salads!'
        )),
    'son <= daughter, dad, mom':
        (sons_entry, [daughters_entry, dads_entry, moms_entry], CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Bring your football and frisbee!'
        )),
    'daughter <= dad, mom, son':
        (sons_entry, [daughters_entry, dads_entry, moms_entry], CalendarEvent(
            title='Family Picnic',
            day=date(2025, 6, 20),
            time='11:00 AM',
            location='Central Park',
            description='Annual family picnic with games and BBQ.'
        )),
})
def test_apply__merge_set(baseline: CalendarEvent, target: list[CalendarEvent], expected: CalendarEvent):
    change_set = MergeSet(baseline.model_copy(), *target, **merge_rules)
    change_set.resolve_preferences()
    assert change_set.apply() == expected


@labeled_tests({
    '1 other model':
        (dads_entry, [moms_entry], {
            'time': ('11:00 AM', [
                '12:00 PM'
            ]),
            'description': ('Annual family picnic with games and BBQ.', [
                'Picnic with family and friends, do not forget the salads!'
            ])
        }),
    'multiple other models': [
        (dads_entry, [moms_entry, sons_entry], {
            'day': (date(2025, 6, 20), [
                date(2025, 6, 20),
                date(2025, 6, 19)
            ]),
            'time': ('11:00 AM', [
                '12:00 PM',
                '12:00 PM'
            ]),
            'location': ('Central Park', [
                'Central Park',
                None
            ]),
            'description': ('Annual family picnic with games and BBQ.', [
                'Picnic with family and friends, do not forget the salads!',
                'Bring your football and frisbee!'
            ])
        }),
        (dads_entry, [moms_entry, daughters_entry], {
            'time': ('11:00 AM', [
                '12:00 PM',
                'All Day'
            ]),
            'location': ('Central Park', [
                'Central Park',
                'The Park'
            ]),
            'description': ('Annual family picnic with games and BBQ.', [
                'Picnic with family and friends, do not forget the salads!',
                None
            ])
        }),
        (dads_entry, [sons_entry, daughters_entry], {
            'day': (date(2025, 6, 20), [
                date(2025, 6, 19),
                date(2025, 6, 20)
            ]),
            'time': ('11:00 AM', [
                '12:00 PM',
                'All Day',
            ]),
            'location': ('Central Park', [
                None,
                'The Park'
            ]),
            'description': ('Annual family picnic with games and BBQ.', [
                'Bring your football and frisbee!',
                None
            ])
        }),
    ],
    'model order': [
        (dads_entry, [moms_entry, sons_entry, daughters_entry], {
            'day': (date(2025, 6, 20), [
                date(2025, 6, 20),
                date(2025, 6, 19),
                date(2025, 6, 20)
            ]),
            'time': ('11:00 AM', [
                '12:00 PM',
                '12:00 PM',
                'All Day'
            ]),
            'location': ('Central Park', [
                'Central Park',
                None,
                'The Park'
            ]),
            'description': ('Annual family picnic with games and BBQ.', [
                'Picnic with family and friends, do not forget the salads!',
                'Bring your football and frisbee!',
                None
            ])
        }),
        (sons_entry, [daughters_entry, moms_entry, dads_entry], {
            'day': (date(2025, 6, 19), [
                date(2025, 6, 20),
                date(2025, 6, 20),
                date(2025, 6, 20)
            ]),
            'time': ('12:00 PM', [
                'All Day',
                '12:00 PM',
                '11:00 AM'
            ]),
            'location': (None, [
                'The Park',
                'Central Park',
                'Central Park'
            ]),
            'description': ('Bring your football and frisbee!', [
                None,
                'Picnic with family and friends, do not forget the salads!',
                'Annual family picnic with games and BBQ.'
            ])
        }),
        (moms_entry, [dads_entry, daughters_entry, sons_entry], {
            'day': (date(2025, 6, 20), [
                date(2025, 6, 20),
                date(2025, 6, 20),
                date(2025, 6, 19)
            ]),
            'time': ('12:00 PM', [
                '11:00 AM',
                'All Day',
                '12:00 PM'
            ]),
            'location': ('Central Park', [
                'Central Park',
                'The Park',
                None,
            ]),
            'description': ('Picnic with family and friends, do not forget the salads!', [
                'Annual family picnic with games and BBQ.',
                None,
                'Bring your football and frisbee!',
            ])
        }),
        (daughters_entry, [moms_entry, dads_entry, sons_entry], {
            'day': (date(2025, 6, 20), [
                date(2025, 6, 20),
                date(2025, 6, 20),
                date(2025, 6, 19)
            ]),
            'time': ('All Day', [
                '12:00 PM',
                '11:00 AM',
                '12:00 PM'
            ]),
            'location': ('The Park', [
                'Central Park',
                'Central Park',
                None
            ]),
            'description': (None, [
                'Picnic with family and friends, do not forget the salads!',
                'Annual family picnic with games and BBQ.',
                'Bring your football and frisbee!'
            ])
        })
    ]
})
def test_merge_set(baseline: CalendarEvent, others: list[CalendarEvent, ...], expected: dict[str, tuple[Any, list[Any]]]):
    assert MergeSet(baseline, *others) == expected


@labeled_tests(get_preferred_tests)
def test_get_preferred(baseline: CalendarEvent, target: CalendarEvent, field: str, expected: Preference):
    assert MergeSet(baseline, *[target] * 3).get_preferred(field) == expected



from typing import Any

from daomodel import DAOModel
from daomodel.model_diff import ModelDiff
from daomodel.testing import labeled_tests
from tests.test_daomodel import SimpleModel, ComplicatedModel, simple_instance, complicated_instance


@labeled_tests({
    'no diff': [
        (simple_instance, simple_instance, {}),
        (complicated_instance, complicated_instance, {}),
        (complicated_instance, ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32), {})
    ],
    'single column': [
        (
                simple_instance,
                SimpleModel(pkA=24),
                {'pkA': (23, 24)}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=18, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32),
                {'pkC': (17, 18)}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=77, prop1='prop', prop2='erty', fkA=23, fkB=32),
                {'pkD': (76, 77)}
        )
    ],
    'multiple columns': [
        (
                complicated_instance,
                ComplicatedModel(pkC=18, pkD=86, prop1='prop', prop2='erty', fkA=23, fkB=32),
                {'pkC': (17, 18), 'pkD': (76, 86)}
        )
    ]
})
def test_model_diff__init(model: DAOModel, other: DAOModel, expected: dict[str, tuple[Any, Any]]):
    assert ModelDiff(model, other, include_pk=True) == expected


@labeled_tests({
    'no diff': [
        (complicated_instance, complicated_instance, {}),
        (complicated_instance, ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32), {}),
        (complicated_instance, ComplicatedModel(pkC=18, pkD=77, prop1='prop', prop2='erty', fkA=23, fkB=32), {})
    ],
    'single column': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2='erty', fkA=23, fkB=32),
                {'prop1': ('prop', 'new')}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='new', fkA=23, fkB=32),
                {'prop2': ('erty', 'new')}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=24, fkB=32),
                {'fkA': (23, 24)}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=33),
                {'fkB': (32, 33)}
        )
    ],
    'multiple columns': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2='prop', fkA=23, fkB=32),
                {'prop1': ('prop', 'new'), 'prop2': ('erty', 'prop')}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=0, fkB=0),
                {'fkA': (23, 0), 'fkB': (32, 0)}
        )
    ],
    'none values': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2=None, fkA=23, fkB=32),
                {'prop1': ('prop', 'new'), 'prop2': ('erty', None)}
        ), (
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2=None, fkA=23, fkB=32),
                complicated_instance,
                {'prop1': ('new', 'prop'), 'prop2': (None, 'erty')}
        )
    ],
    'unset values': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', fkA=23, fkB=32),
                {'prop1': ('prop', 'new'), 'prop2': ('erty', None)}
        ), (
                ComplicatedModel(pkC=17, pkD=76, prop1='new', fkA=23, fkB=32),
                complicated_instance,
                {'prop1': ('new', 'prop'), 'prop2': (None, 'erty')}
        )
    ]
})
def test_model_diff__init__exclude_pk(model: DAOModel, other: DAOModel, expected: dict[str, tuple[Any, Any]]):
    assert ModelDiff(model, other, include_pk=False) == expected



from datetime import date
from typing import Optional

import pytest
from sqlalchemy import ForeignKeyConstraint
from sqlmodel import Field

from daomodel import DAOModel
from daomodel.dao import DAO
from daomodel.fields import Identifier, Unsearchable, ReferenceTo
from daomodel.testing import TestDAOFactory
from daomodel.util import next_id


class Person(DAOModel, table=True):
    name: Identifier[str]
    age: Identifier[int]
    ssn: Unsearchable[Optional[str]] = Field(unique=True)


class Book(DAOModel, table=True):
    name: Identifier[str]
    subject: str
    owner: int = ReferenceTo('student.id')


class Hall(DAOModel, table=True):
    location: Identifier[str]
    floor: Identifier[int]
    color: str


class Locker(DAOModel, table=True):
    __table_args__ = (
        ForeignKeyConstraint(
            ['location', 'floor'],
            ['hall.location', 'hall.floor'],
        ),
    )

    number: Identifier[int]
    owner: int = ReferenceTo('student.id')
    location: str
    floor: int


class BasePerson(DAOModel):
    id: Identifier[int]
    name: Optional[str]


class Staff(BasePerson, table=True):
    hire_date: date


class Student(BasePerson, table=True):
    gender: Optional[str]
    active: bool = True

    class Meta:
        searchable_relations = {
            Book.name, Book.subject, Locker.number,
            (Locker, Hall.color)
        }


@pytest.fixture(name='person_dao')
def person_dao_fixture(daos: TestDAOFactory) -> DAO:
    dao = daos[Person]
    dao.create_with(name='John', age=23)
    dao.create_with(name='Joe', age=32)
    dao.create_with(name='Mike', age=25)
    dao.create_with(name='John', age=45)
    dao.create_with(name='Greg', age=31)
    dao.create_with(name='Mike', age=18)
    dao.create_with(name='Paul', age=25)
    return dao


pk_ordered = [
    Person(name='Greg', age=31),
    Person(name='Joe', age=32),
    Person(name='John', age=23),
    Person(name='John', age=45),
    Person(name='Mike', age=18),
    Person(name='Mike', age=25),
    Person(name='Paul', age=25)
]
age_ordered = [
    Person(name='Mike', age=18),
    Person(name='John', age=23),
    Person(name='Mike', age=25),
    Person(name='Paul', age=25),
    Person(name='Greg', age=31),
    Person(name='Joe', age=32),
    Person(name='John', age=45)
]
duplicated_names = [
    Person(name='John', age=23),
    Person(name='John', age=45),
    Person(name='Mike', age=18),
    Person(name='Mike', age=25)
]
unique_names = [
    Person(name='Greg', age=31),
    Person(name='Joe', age=32),
    Person(name='Paul', age=25)
]


def setup_students(daos: TestDAOFactory) -> DAO:
    dao = daos[Student]
    dao.create_with(id=100, name='Bob', gender='m')
    dao.create_with(id=next_id(), name='Mike', gender='m')
    dao.create_with(id=next_id(), name='John', gender='m', active=False)
    dao.create_with(id=next_id(), name='Jerry', gender='m')
    dao.create_with(id=next_id(), name='Carl', gender='m')
    dao.create_with(id=next_id(), name='Barbara', gender='f')
    dao.create_with(id=next_id(), name='Mary', gender='f', active=False)
    dao.create_with(id=next_id(), name='Jill', gender='f', active=False)
    dao.create_with(id=next_id(), name='Jessie', gender='f')
    dao.create_with(id=next_id(), name='Karen', gender='f')
    dao.create_with(id=next_id(), name='Sam')
    dao.create_with(id=next_id(), active=False)
    dao.create(next_id())
    return dao


@pytest.fixture(name='student_dao')
def student_dao_fixture(daos: TestDAOFactory) -> DAO:
    return setup_students(daos)


@pytest.fixture(name='school_dao')
def school_dao_fixture(daos: TestDAOFactory) -> DAO:
    dao = setup_students(daos)
    book_dao = daos[Book]
    book_dao.create_with(name='Biology 101', subject='Science', owner=100)
    book_dao.create_with(name='Art of the World', subject='Art', owner=101)
    book_dao.create_with(name='Pre-Calculus', subject='Math', owner=102)
    book_dao.create_with(name='Calculus', subject='Math', owner=103)
    hall_dao = daos[Hall]
    hall_dao.create_with(location='EAST', floor=1, color='green')
    hall_dao.create_with(location='WEST', floor=1, color='red')
    hall_dao.create_with(location='EAST', floor=2, color='blue')
    hall_dao.create_with(location='WEST', floor=2, color='yellow')
    locker_dao = daos[Locker]
    locker_dao.create_with(number=1101, owner=100, location='EAST', floor=1)
    locker_dao.create_with(number=1102, owner=108, location='EAST', floor=1)
    locker_dao.create_with(number=1103, owner=103, location='EAST', floor=1)
    locker_dao.create_with(number=1104, owner=104, location='WEST', floor=1)
    locker_dao.create_with(number=1105, owner=109, location='WEST', floor=1)
    locker_dao.create_with(number=1106, owner=101, location='WEST', floor=1)
    locker_dao.create_with(number=1201, owner=107, location='EAST', floor=2)
    locker_dao.create_with(number=1202, owner=110, location='EAST', floor=2)
    locker_dao.create_with(number=1203, owner=102, location='EAST', floor=2)
    locker_dao.create_with(number=1204, owner=106, location='WEST', floor=2)
    locker_dao.create_with(number=1205, owner=111, location='WEST', floor=2)
    locker_dao.create_with(number=1206, owner=105, location='WEST', floor=2)
    return dao


def split(to_split: list, target_size: int) -> tuple[list, ...]:
    return tuple(to_split[x:x+target_size] for x in range(0, len(to_split), target_size))


all_students = [
    Student(id=100),
    Student(id=101),
    Student(id=102),
    Student(id=103),
    Student(id=104),
    Student(id=105),
    Student(id=106),
    Student(id=107),
    Student(id=108),
    Student(id=109),
    Student(id=110),
    Student(id=111),
    Student(id=112)
]
active = [
    Student(id=100),
    Student(id=101),
    Student(id=103),
    Student(id=104),
    Student(id=105),
    Student(id=108),
    Student(id=109),
    Student(id=110),
    Student(id=112)
]
inactive = [
    Student(id=102),
    Student(id=106),
    Student(id=107),
    Student(id=111)
]
active_females = [
    Student(id=105),
    Student(id=108),
    Student(id=109)
]
having_gender = [
    Student(id=100),
    Student(id=101),
    Student(id=102),
    Student(id=103),
    Student(id=104),
    Student(id=105),
    Student(id=106),
    Student(id=107),
    Student(id=108),
    Student(id=109)
]
having_book = [
    Student(id=100),
    Student(id=101),
    Student(id=102),
    Student(id=103)
]
having_math_book = [
    Student(id=102),
    Student(id=103)
]
not_having_name = [
    Student(id=111),
    Student(id=112)
]
not_having_locker = [
    Student(id=112)
]
in_blue_hall = [
    Student(id=102),
    Student(id=107),
    Student(id=110)
]
page_one, page_two, page_three = split(all_students, 5)
